from typing import List, Tuple, Union
import numpy as np
from scipy.sparse import hstack, vstack, lil_array
import itertools
from collections import Counter
import re
from nltk.corpus import stopwords
from scipy.sparse import csr_array, find
from nltk import WordNetLemmatizer
from HanTa import HanoverTagger as ht
from tqdm import tqdm
lemma = WordNetLemmatizer()

def preprocess(texts, language="english", individual_stop_word_list=None, verbose=False):
    """
    Implements a very barebone preprocessing procedure for english texts. Is used by the pipeline-method and can be replaced by any arbitrary preprocessing function.
    Args:
        texts: list of texts
        language: language of the texts
        individual_stop_word_list: list of stop words, optional
        verbose: boolean, optional
    Returns:
        list of preprocessed texts
    """
    if not isinstance(language, str):
        raise TypeError("language must be a string!")
    if not isinstance(texts, list):
        try:
            texts = list(texts)
        except ValueError:
            raise TypeError("texts must be a list of strings!")
    if not isinstance(texts[0], str):
        raise TypeError("texts must be a list of strings!")
    if individual_stop_word_list is not None:
        if not isinstance(individual_stop_word_list, list):
            try:
                individual_stop_word_list = list(individual_stop_word_list)
            except ValueError:
                raise TypeError("individual_stop_word_list must be a list of strings!")
        if not isinstance(individual_stop_word_list[0], str):
            raise TypeError("individual_stop_word_list must be a list of strings!")
    if not isinstance(verbose, bool):
        raise TypeError("verbose must be a boolean!")
    if language == "german":
        tagger = ht.HanoverTagger('morphmodel_ger.pgz')
    elif language == "dutch":
        tagger = ht.HanoverTagger('morphmodel_dutch.pgz')
    elif language == "english":
        tagger = ht.HanoverTagger('morphmodel_en.pgz')
    else:
        tagger = None
    processed_texts = []
    stop = stopwords.words(language) if individual_stop_word_list is None else individual_stop_word_list
    stop = set([re.sub(r"[^a-zäöüß ]", "", x) for x in stop])
    iterator = tqdm(texts) if verbose else texts
    for text in iterator:
        text = re.sub(r"[\s-]+", " ", text)
        text = re.sub(r"[^a-zäöüßA-ZÄÖÜ ]", "", text).split()
        if tagger is not None:
            tagged_text = tagger.tag_sent(text)
            text = [x[1].lower() for x in tagged_text]
        else:
            text = [lemma.lemmatize(x).lower() for x in text]
        text = [x for x in text if x not in stop and len(x) > 2]
        processed_texts.append(text)
    return processed_texts

def create_dtm(texts: List[List[str]], vocab: List[str], min_count: int = 5, deleted_indices=None, dtm: np.ndarray = None) -> Tuple[np.ndarray, List[str], List[int]]:
    """
    Creates a document-term matrix from a list of texts and updates the existing dtm if there is one.
    Stores both the dtm and the vocabulary in the class.
    Args:
        texts: list of texts
    Returns:
        None
    """
    if not isinstance(texts, list):
        try:
            texts = list(texts)
        except ValueError:
            raise TypeError("texts must be a list of list of strings!")
    if not isinstance(texts[0], list):
        try:
            texts = [list(x) for x in texts]
        except ValueError:
            raise TypeError("texts must be a list of lists of strings!")
    if not isinstance(texts[0][0], str):
        try:
            texts = [[str(x) for x in text] for text in texts]
        except ValueError:
            raise TypeError("texts must be a list of lists of strings!")
    if not isinstance(vocab, list):
        try:
            vocab = list(vocab)
        except ValueError:
            raise TypeError("vocab must be a set of strings!")
    if not isinstance(min_count, int):
        try:
            min_count = int(min_count)
        except ValueError:
            raise TypeError("min_count must be an integer!")
    if not deleted_indices:
        deleted_indices = []
    elif not isinstance(deleted_indices, list):
        raise TypeError("deleted_indices must be a list of integers!")
    elif not isinstance(deleted_indices[0], int) and not isinstance(deleted_indices[0], np.int64):
        raise TypeError("deleted_indices must be a list of integers!")
    if dtm is not None:
        if not isinstance(dtm, np.ndarray) and not isinstance(dtm, csr_array):
            raise TypeError("dtm must be a numpy array!")

    counters = [Counter(doc) for doc in texts]
    all_new_words_counted = Counter()
    for obj in counters:
        all_new_words_counted.update(obj)
    # all_new_words_counted = Counter([word for doc in texts for word in doc])
    old_vocab_set = set(vocab)
    new_vocabulary = vocab + list(set([word for word, value in all_new_words_counted.items() if value >= min_count and word not in old_vocab_set]))
    new_vocabulary_index = {word: i for i, word in enumerate(new_vocabulary)}
    new_vocabulary_set = set(new_vocabulary)

    # non_deleted_bool = [len([word for word in text if word in new_vocabulary_set]) > 0 for i, text in enumerate(texts)]
    # non_deleted_indices = np.argwhere(non_deleted_bool).flatten()
    # deleted_indices.extend(non_deleted_indices + max(deleted_indices + [0]))
    # texts = list(itertools.compress(texts, non_deleted_bool))

    updated_dtm = lil_array((len(texts), len(new_vocabulary)))
    for i, doc in enumerate(texts):
        updated_dtm[i, [new_vocabulary_index[word] for word in counters[i].keys() if word in new_vocabulary_set]] = \
            [value for word, value in counters[i].items() if word in new_vocabulary_set]

    number_of_new_words = len(new_vocabulary) - len(vocab)
    vocab = new_vocabulary
    if dtm is not None:
        new_data = lil_array((dtm.shape[0], number_of_new_words))
        existing_dtm = hstack((dtm, new_data))
        combined_dtm = vstack((existing_dtm, updated_dtm))
        dtm = combined_dtm.tocsr()
    else:
        dtm = updated_dtm.tocsr()
    return dtm, vocab, deleted_indices

def get_word_and_doc_vector(dtm: Union[csr_array, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Turns a document-term matrix into index vectors. The word vector contains the vocabulary index for each word
    occurrence including multiple occurrences in one text. The document vector contains the document index for each
    word occurrence including multiple occurrences in one document.
    Args:
        dtm: document-term matrix
    Returns:
        word_vec: word vector
        doc_vec: document vector
    """
    if not isinstance(dtm, csr_array) or isinstance(dtm, np.ndarray):
        try:
            dtm = np.array(dtm)
            if len(dtm.shape) == 0:
                raise ValueError
        except ValueError:
            raise TypeError("dtm must be a numpy array!")
    if dtm.shape[0] == 0 or dtm.shape[1] == 0:
        raise ValueError("dtm must not be empty!")
    (row_nonzero, column_nonzero) = dtm.nonzero()
    _, _, value_list = find(dtm)
    value_list = value_list.astype(int)

    word_vec = []
    doc_vec = []
    for i, elem in enumerate(row_nonzero):
        word_vec += [column_nonzero[i]] * value_list[i]
        doc_vec += [elem] * value_list[i]
    return np.array(word_vec, dtype=np.uint64), np.array(doc_vec, dtype=np.uint64)




import json
import os
import time
import logging
from typing import List, Dict
from ttta.preprocessing.semantic import OxfordDictAPI
from ttta.preprocessing.settings import EmbeddingFiles
from ttta.preprocessing.schemas import Words
from pydantic import parse_obj_as


def get_words_and_senses(path:str) -> List[Dict]:
    """
    Get all the words and their senses from the Oxford API, given a path to a file containing the words
    Args:
        path: str
            Path to the file containing the words
    Returns:
        List[Dict]: A list of dictionaries containing the words and their senses

    """
    all_words = []
    if not os.path.exists(path):
        raise ValueError(
            f'No matches for the given path: {path}'
        )
    with open(path) as f: full_text = f.read()
    for idx, word in enumerate(full_text.split('\n')):
        try:
            all_words += [OxfordDictAPI(word_id=word).get_senses()]
            if (idx != 0) and (idx % 10 == 0):
                logging.info('Waiting for 3 seconds before continuing...')
                time.sleep(3)
        except ValueError:
            continue

    if len(all_words) != len(full_text.split('\n')):
        logging.warning(
            f'Number of words processed: {len(all_words)} is inferior to the number of words in the file: {path}')

    return all_words


def poly_words(oxford_words_file: str):
    """
    Get all words that have more than one sense
    Args:
        oxford_words:

    Returns:
        str: the words that have more than one sense
    """
    with open(oxford_words_file) as f: oxford_words = parse_obj_as(List[Words], json.load(f))
    for word in oxford_words:
        if len(word.senses) > 1:
            yield word.word



if __name__ == '__main__':
    files = EmbeddingFiles()

    with open(files.oxford_word_senses, 'w') as f:
        p = get_words_and_senses(files.poly_words_f)
        json.dump(p, f, indent=4)




import numpy as np

def calculate_importance(word_topic_matrix: np.ndarray) -> np.ndarray:
    """
    Calculate the importance of words for each topic, weighting down words frequently used in multiple topics
    Args:
        word_topic_matrix: word-topic matrix
    Returns:
        importance: importance of words for each topic
    """
    if not isinstance(word_topic_matrix, np.ndarray):
        raise TypeError("word_topic_matrix must be a numpy array!")
    word_topic_matrix = word_topic_matrix.transpose()
    denominator = np.sum(word_topic_matrix, axis=1)[:, np.newaxis]
    denominator[denominator == 0] = 1
    importance = word_topic_matrix / denominator
    log_importance = np.log(importance + 1e-5)
    importance = importance * (log_importance - np.mean(log_importance, axis=0)[np.newaxis, :])
    return importance



import pandas as pd
from sklearn.base import BaseEstimator
import numpy as np
from sklearn.utils.extmath import randomized_svd
from pathlib import Path
import pickle
from tqdm.autonotebook import trange, tqdm
from typing import NamedTuple
import numpy.typing as npt
from scipy.optimize import minimize
from ..preprocessing.preprocess import create_dtm


class PoissonReducedRankParameters(NamedTuple):
    """Parameter class used for storing the parameters of the poisson reduced rank model.
    Given a tdm of dimensions (J_tokens, L_documents), the dimensions of the other values are described as follows
    alpha has dimensions (J_tokens,)
    beta has dimensions (L_documents,)
    b has dimensions (J_tokens, K) with K being equal to the dimensions of the political space
    f has diemensions (K, L_documents) with K being equal to the dimensions of the political space
    """

    logged_dtm: np.ndarray
    alpha: npt.NDArray[np.float32]
    beta: npt.NDArray[np.float32]
    b: np.ndarray
    f: np.ndarray
    svd_u: np.ndarray
    svd_v: np.ndarray
    svd_s: np.ndarray


def poisson_log_likelihood(Y: np.ndarray, mu: np.ndarray) -> np.float32:
    """Computes the log likelihood over two arrays. The implementation ignores the factorial of y since this has no effect on the maximization and can be ommitted for optimization.

    Args:
        Y (np.ndarray): Matrix or vector of y values (in this package: poisson distributed word counts)
        mu (np.ndarray): mean parameter matrix of poisson parameter

    Returns:
        np.double: returns the sum of the log likelihood
    """
    assert Y.shape == mu.shape, "Y, and theta must be of same shape"
    return np.sum(-mu + np.multiply(Y, np.log(mu)))


def poisson_log_likelihood_parameterized(
    y: np.ndarray, alpha: np.ndarray, beta: np.ndarray, b: np.ndarray, f: np.ndarray
) -> np.float32:
    """Paramterizes the poisson log likelihood for the model Poisson Reduced Rank regression with time independent word weights in order to wrap these functions for the scipy optimization.

    Args:
        y (np.ndarray): vector of token counts (n_tokens)
        alpha (np.ndarray): vector of average word_weights (n_tokens)
        beta (np.ndarray): vector of average counts per document (m_doucuments)
        b (np.ndarray): vector of word weights for the j-th token (n_tokens, k_latent space dimensions)
        f (np.ndarray): vector of l-th document specific weights

    Returns:
        np.float32: poisson log likelihood value
    """
    if f.ndim == 2 and b.ndim == 2:
        bf = np.dot(b, f)
        logged_mu = bf + alpha.reshape(-1, 1) + beta
    else:
        bf = np.multiply(b, f)
        logged_mu = bf + alpha + beta  # formula for the logged likelihood
    pois_loglik_calc = poisson_log_likelihood(y, np.exp(logged_mu))
    return pois_loglik_calc


class PoissonReducedRankTimeIndependentWordWeights(BaseEstimator):

    def __init__(self, K=2):
        self.k = K
        self._L_documents = 0
        self._J_tokens = 0
        self.data = None

    def _df2dtm(
        self,
        texts: pd.DataFrame,
        text_column: str = "text",
        date_column: str = "date",
        individual_column: str = "individual",
    ) -> np.ndarray:
        """Converts the input pandas data frame to a document term frequency matrix.


        Args:
            texts (pd.DataFrame): A pandas DataFrame containing the columns text_column, date_column and individual_column containing the documents and their respective dates. The column combination of individual_column and date_column leads to a panel like data structure.
                   The dates must be in a format interpretable by pandas.to_datetime(). Each element of the text_column column must be a list of strings. Internally, the create_dtm function is used with no control over the vocab argument of the create_dtm function currently.
            text_column (str): The name of the column in texts containing the documents. Defaults to "text".
            date_column (str): The name of the column in texts containing the dates. Defaults to "date".
            individual_column (str): An id column, containing the ids (ints) of the individual. Defaults to "individual".
        Returns:
            np.ndarray: Returns an np.ndarray for which can be used for the fitting
        """
        if text_column not in texts.columns:
            raise ValueError(
                "Column name given by text_column must be present in data frame columns"
            )
        if date_column not in texts.columns:
            raise ValueError(
                "Column name given by date_column must be present in data frame columns"
            )
        if individual_column not in texts.columns:
            raise ValueError(
                "Column name given by individual_column must be present in data frame columns"
            )
        ## Data in the text_column is unpacked and sorted
        data: pd.DataFrame = texts.copy()
        data[date_column] = pd.to_datetime(data[date_column])
        data.sort_values([individual_column, date_column], inplace=True)
        # add missing document, time rows and fill rows with 0 of token matrix
        texts_list = sum(data[text_column].to_list(), [])
        vocab = list(dict.fromkeys(texts_list))
        ## create dtm from all given texts, since the create_dtm function takes a vocab, currently, all present strings are used as the vocab
        dtm_data = create_dtm(texts=data.loc[:, text_column], vocab=vocab)
        dtm = dtm_data[0]  # extracts dtm matrix from result
        ## convert dtm to dataframe
        df_dtm = pd.DataFrame(dtm.toarray())
        df_dtm.columns = [f"feature_{i}" for i in range(df_dtm.shape[1])]
        df_dtm.set_index(
            data.index, inplace=True
        )  # reset index of the dtm_matrix to match it to the initial dataframe
        ## add dtm matrix columns to dataframe
        data = pd.concat([data, df_dtm], axis=1)
        ## add missing rows for (time, individual) dimension
        missing_rows_added = (
            data.set_index([individual_column, date_column])
            .unstack(fill_value=0)
            .stack()
            .reset_index()
        )
        columns = ["feature" in x for x in missing_rows_added.columns]
        matrix = missing_rows_added.loc[:, columns].transpose()

        ## save data for later viewing in class
        self.data = missing_rows_added.copy()

        return matrix.to_numpy()

    @staticmethod
    def _normalize_matrix(X: npt.NDArray) -> dict:
        """Normalizes a matrix and returns a normalized matrix namedTuple, containing average occurences of tokens per token across documents and per document across token, as well as the centered document term matrix.

        Args:
            X (np.ndarray): matrix of the form (n_tokens, n_documents)

        Returns:
            dict: dictionary containing stats for the normalized matrix
        """
        mean_occurence_per_token = np.mean(
            X, axis=1
        )  # mean occurence of tokens (corresponds to alpha in formula)
        # mean occurence of each token in document
        mean_occurence_per_document = np.mean(
            X, axis=0
        )  # avg. token occurence per document
        matrix_mean = np.mean(X)

        dtm_row_normalized = np.subtract(X, mean_occurence_per_token.reshape(-1, 1))
        dtm_column_normalized = np.subtract(
            dtm_row_normalized, mean_occurence_per_document
        )

        return {
            "centered_matrix": dtm_column_normalized,
            "matrix_mean": matrix_mean,
            "mean_per_row": mean_occurence_per_token,
            "mean_per_column": mean_occurence_per_document,
        }

    def _init_params(self):
        # TODO: rewrite the methods with an internal method
        self.alpha = np.empty(self._J_tokens)
        self.beta = np.empty(self._L_documents)
        self.b = np.empty((self._J_tokens, self.k))
        self.f = np.empty((self._L_documents, self.k))

    def get_params(self) -> dict:
        """Returns the parameters for the poisson reduced rank model with independent word weights

        Returns:
            dict: Dictionary containing the parameter arrays
        """
        return {
            "K": self.k,
            "alpha": self.alpha,
            "beta": self.beta,
            "b": self.b,
            "f": self.f,
        }

    def get_random_params(self) -> dict:
        """Returns random parameters for the poisson reduced rank model with independent word weights

        Returns:
            dict: Dictionary containing the random paramters
        """
        return {
            "K": self.k,
            "alpha": np.random.rand(
                self._J_tokens,
            ),
            "beta": np.random.rand(
                self._L_documents,
            ),
            "b": np.random.rand(self._J_tokens, self.k),
            "f": np.random.rand(self._L_documents, self.k),
        }

    def set_params(self, **params) -> None:
        """sets the internal parameters of the poisson reduced rank model. The **params should contains 'alpha', 'beta', 'b' and 'f'. 'K' is not allowed, since it would change the whole metho due to the increase in dimensions.
        The dimensionalities are not checked, please make sure that the np.ndarrays are of the correct shape.
        """

        if "alpha" in params:
            self.alpha = params["alpha"]

        if "beta" in params:
            self.beta = params["beta"]

        if "b" in params:
            self.b = params["b"]

        if "f" in params:
            self.f = params["f"]

    def save(self, path: Path) -> None:
        """Saves the Poisson Reduced rank model with independent word weights to the given path
        Args:
            path (Path): Path object to the given save location
        """
        with open(path.absolute(), "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: Path) -> dict:
        with open(path.absolute(), "rb") as f:
            return pickle.load(f)

    def fit(
        self,
        texts: pd.DataFrame,
        text_column: str = "text",
        date_column: str = "date",
        individual_column: str = "individual",
        n_iter: int = 100,
        tol: float = 1e-5,
    ):
        """Fits the Poisson Reduced Rank Model with independent word weights to the the given data frame. Internally, the given pandas dataframe is converted to a numpy document term frequency matrix.

        Args:
            texts (pd.DataFrame): A pandas DataFrame containing the columns text_column, date_column and individual_column containing the documents and their respective dates. The column combination of individual_column and date_column leads to a panel like data structure.
                   The dates must be in a format interpretable by pandas.to_datetime(). Each element of the text_column column must be a list of strings. Internally, the create_dtm function is used with no control over the vocab argument of the create_dtm function currently.
            text_column (str): The name of the column in texts containing the documents. Defaults to "text".
            date_column (str): The name of the column in texts containing the dates. Defaults to "date".
            individual_column (str): An id column, containing the ids (ints) of the individual. Defaults to "individual".
            n_iter (int, optional): number of iteration for fitting. Defaults to 100.
            tol (float, optional): Currently not implemented. Defaults to 1e-5.

        Returns:
            None: Returns nothing. To acces the final parameters, please see the get_params() method
        """
        if self.k <= 0:
            raise ValueError("Dimensions K must be greater 0")
        if n_iter <= 0:
            raise ValueError("Number of iterations must be greater 0")
        print("Converting Dataframe to dtm matrix...")
        ## convert the pandas dataframe to a dtm matrix
        X = self._df2dtm(
            texts=texts,
            text_column=text_column,
            date_column=date_column,
            individual_column=individual_column,
        )  # X (npt.NDArray[np.int32]): array-like of shape (n_tokens, n_documents).
        print("Conversion from Dataframe to dtm successful!")

        ## initialize basic dimensions from given data
        self._J_tokens, self._L_documents = X.shape[0], X.shape[1]

        print(
            f"Fitting a poisson Reduced Rank Model with time indep. Word Weights for {self._J_tokens} tokens and {self._L_documents} documents"
        )
        self._init_params()

        model_params = self._calculate_starting_values(X)

        for i in trange(n_iter):
            log_likelihood_old = poisson_log_likelihood_parameterized(
                X,
                alpha=model_params.alpha,
                beta=model_params.beta,
                b=model_params.b,
                f=model_params.f,
            )
            update_f_and_beta = self._update_f_and_beta(X, model_params)

            updated_params = PoissonReducedRankParameters(
                logged_dtm=model_params.logged_dtm,
                svd_s=model_params.svd_s,
                svd_u=model_params.svd_u,
                svd_v=model_params.svd_v,
                f=update_f_and_beta.f,
                beta=update_f_and_beta.beta,
                b=model_params.b,
                alpha=model_params.alpha,
            )
            update_b_and_alpha = self._update_b_and_alpha(X, updated_params)

            model_params = update_b_and_alpha  # contains updated parameters
            log_likelihood_new = poisson_log_likelihood_parameterized(
                X,
                alpha=model_params.alpha,
                beta=model_params.beta,
                b=model_params.b,
                f=model_params.f,
            )

            if np.abs(log_likelihood_new - log_likelihood_old) < tol:
                print(
                    "Aborted further updates since the Early Stopping (likelihood) executed."
                )
                break

        reparameterized_params = self._reparameterize_identification(model_params)
        return reparameterized_params

    def _reparameterize_identification(
        self, params: PoissonReducedRankParameters
    ) -> PoissonReducedRankParameters:
        """Reparameterizes the parameters in order to meet the identification conditions of (2) and (3) given in the paper.

        Args:
            params (PoissonReducedRankParameters): Current parameters, of the model

        Returns:
            PoissonReducedRankParameters: Reparameterized parameters which fulfill the identification conditions.
        """

        # get the logged m_jl
        log_mu_jl = self._convert_parameters2loggedPoissonParameters(params)
        # How to parameterize centered vector: https://mc-stan.org/docs/2_18/stan-users-guide/parameterizing-centered-vectors.html

        # calculate centered dtm statistics, for svd and calculation of poisson parametization
        centered_dtm_stats = self._normalize_matrix(log_mu_jl)
        alpha = centered_dtm_stats["mean_per_row"]
        beta = centered_dtm_stats["mean_per_column"] - centered_dtm_stats["matrix_mean"]

        # next, we need to calculate b_k and f_k, where k defines the dimension of the political space
        # this is done by decomposing the centered dtm
        centered_theta = np.subtract(np.subtract(log_mu_jl, alpha.reshape(-1, 1)), beta)
        u, s, v = self._wrapper_preset_svd(centered_theta)

        # b_k corresponds to the k-th column
        b = np.divide(np.multiply(s, u), np.sqrt(self._L_documents))
        # f_k corresponds to the k-th column
        f = np.multiply(np.sqrt(self._L_documents), v)

        return PoissonReducedRankParameters(
            logged_dtm=centered_theta,
            alpha=alpha,
            beta=beta,
            f=f,
            b=b,
            svd_v=v,
            svd_u=u,
            svd_s=s,
        )

    def _update_f_and_beta(
        self, X: np.ndarray, model_params: PoissonReducedRankParameters
    ) -> PoissonReducedRankParameters:
        """Updates the f matrix of the poisson reduced rank model, given the other vectors/matrices (alpha, beta and b).
        As stated in the supplementary material of the paper, the maximization of the matrix is identical to seperatly maximizing the matrices column wise.


        Args:
            Y (np.ndarray): logged word occurence
            logged_param_mat (np.ndarray): Logged parameter matrix of the poisson parameter
        """

        new_f = np.empty(shape=model_params.f.shape, dtype=np.float32)
        new_beta = np.empty(shape=model_params.beta.shape, dtype=np.float32)

        for i_col in range(self._L_documents):
            # update f (given alpha_j's and b_j's)
            init_guess = np.concatenate(
                [model_params.beta[i_col].reshape(-1), model_params.f[:, i_col]]
            )  # params to optimize
            update_f_and_beta_result = minimize(
                PoissonReducedRankTimeIndependentWordWeights.wrapper_update_f_poisson_log_likelihood,
                init_guess,
                args=(model_params.alpha, model_params.b, X[:, i_col]),
                options={"maxiter": 1},
            )
            new_f[:, i_col] = update_f_and_beta_result.x[1:]
            new_beta[i_col] = update_f_and_beta_result.x[0]

        return PoissonReducedRankParameters(
            logged_dtm=model_params.logged_dtm,
            alpha=model_params.alpha,
            b=model_params.b,
            f=new_f,
            beta=new_beta,
            svd_s=model_params.svd_s,
            svd_v=model_params.svd_v,
            svd_u=model_params.svd_u,
        )

    def _update_b_and_alpha(
        self, X: np.ndarray, model_params: PoissonReducedRankParameters
    ) -> PoissonReducedRankParameters:
        """Updates the parameter vectors b and intercept alpha for each j. The update is done independent of j and given the latent vector f per document as well as the average token count per document beta_l.

        Args:
            X (np.ndarray): The word count matrix of dimensions (J_tokens, L_documents)
            model_params (PoissonReducedRankParameters): Model params named tuple, containing the current parameters.

        Returns:
            PoissonReducedRankParameters: NamedTuple containing the updated current parameters.
        """
        new_b = np.empty(shape=model_params.b.shape, dtype=np.float32)
        new_alpha = np.empty(shape=model_params.alpha.shape, dtype=np.float32)

        for j_token in range(self._J_tokens):
            init_guess = np.concatenate(
                [model_params.alpha[j_token].reshape(-1), model_params.b[j_token, :]]
            )
            update_b_and_alpha_result = minimize(
                PoissonReducedRankTimeIndependentWordWeights.wrapper_update_b_poisson_log_likelihood,
                init_guess,
                args=(model_params.beta, model_params.f, X[j_token, :]),
                options={"maxiter": 1},
            )
            new_b[j_token, :] = update_b_and_alpha_result.x[1:]
            new_alpha[j_token] = update_b_and_alpha_result.x[0]

        new_params = PoissonReducedRankParameters(
            logged_dtm=model_params.logged_dtm,
            alpha=new_alpha,
            b=new_b,
            f=model_params.f,
            beta=model_params.beta,
            svd_s=model_params.svd_s,
            svd_v=model_params.svd_v,
            svd_u=model_params.svd_u,
        )

        return new_params

    def _convert_parameters2loggedPoissonParameters(
        self, parameters: PoissonReducedRankParameters
    ) -> np.ndarray:
        """Uses the calculated parameters to construct a matrix of logged poisson parameters.

        Args:
            parameters (PoissonReducedRankParameters): Parameters for the poisson reduced rank model. I.e a PoissonReducedRankParameters object.

        Returns:
            np.ndarray: Matrix of logged poisson parameters
        """
        # INFO: In the r-package, the data was computed differently, (see estim.R line 84-90). However,
        # the calculation there was not clear for me, hence I followed the formula in the original paper.
        # theta = parameters.svd_u.dot(diag_s).dot(parameters.svd_v)
        # add th_rm -> rowmeans(theta - theta_mean), th_cm, but must be converted to matrix colmean(theta-theta_mean), th_m (scalar)
        bf = np.dot(parameters.b, parameters.f)
        bf_added_alpha = np.add(
            bf, parameters.alpha.reshape(-1, 1)
        )  ## add alpha vector to each column
        logged_mu = np.add(
            bf_added_alpha, np.transpose(parameters.beta)
        )  ## add row vector beta to result from prev. step

        return logged_mu

    @staticmethod
    def is_identified_correctly(params: PoissonReducedRankParameters, atol=0.1) -> bool:
        """Tests if the given parameters, meet the identification conditions (2) & (3) in section 2.1. of "Poisson reduced-rank models with an application to political text data of Jentsch et. al.

        Args:
            params (PoissonReducedRankParameters): Parameters of the fitted model
            atol (float, optional): Absolute difference to the identification result. Defaults to 0.1.

        Returns:
            bool: True, if all identification conditions are fullfilled, otherwise false.
        """
        beta_sum = np.sum(params.beta)
        b_sum = np.sum(params.b, axis=0)
        f_sum = np.sum(params.f, axis=1)

        sum_vector_condition = 0

        # check conditions in (2)
        if np.isclose(beta_sum, [sum_vector_condition], atol=atol) == False:
            print(f"Sum over beta: {beta_sum}")
            return False
        if np.allclose(b_sum, np.zeros(params.b.shape[1]), atol=atol) == False:
            print(f"b sum: {b_sum}")
            return True
        if np.allclose(f_sum, np.zeros(params.f.shape[0]), atol=atol) == False:
            print(f"f sum: {f_sum}")
            return True

        # check conditions in (3)
        # TODO: Implement check

        return True

    def _wrapper_preset_svd(
        self, X: np.ndarray, **kwargs
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Preset singular value decomposition. For more information about the internal method and the result, please see sklearn.utils.extmath.randomized_svd.

        Args:
            X (np.ndarray): np.ndarray (usually 2D matrix) to decompose

        Returns:
            tuple[np.ndarray, np.ndarray,np.ndarray]: (Unitary matrix having left singular vectors with signs flipped as columns. ; The singular values, sorted in non-increasing order. ; Unitary matrix having right singular vectors with signs flipped as rows.

        """
        u, s, v = randomized_svd(
            M=X,
            n_oversamples=10,
            n_iter=2,
            n_components=self.k,
            random_state=None,
            **kwargs,
        )
        return (u, s, v)

    def _calculate_starting_values(
        self, X: npt.NDArray[np.int32]
    ) -> PoissonReducedRankParameters:
        r"""Calculates the starting values for the optimization of the algorithm.
            In particular :math:`\alpha^(0), \beta^(0), b^(0), f^(0)`.
            If log(0) is encountered, a warning is suppresed.

        Args:
            X (npt.NDArray[np.int32]): Document Term Matrix (DTM) of dimensions (n_tokens, n_documents).
        """
        # As the analysis of variance decomposition of log mu_jit gives the true parameters, log x_jit is used to compute the initial values.
        with np.errstate(divide="ignore"):
            theta = np.log(X)
        # replace inifinite and not a number values --> i.e log(0)=-2
        theta[~np.isfinite(theta)] = -2

        # calculate centered dtm statistics, for svd and calculation of poisson parametization
        centered_dtm_stats = self._normalize_matrix(theta)
        alpha = centered_dtm_stats["mean_per_row"]
        beta = centered_dtm_stats["mean_per_column"] - centered_dtm_stats["matrix_mean"]

        # next, we need to calculate b_k and f_k, where k defines the dimension of the political space
        # this is done by decomposing the centered dtm
        centered_theta = theta - alpha.reshape(-1, 1) - beta
        u, s, v = self._wrapper_preset_svd(centered_theta)

        # b_k corresponds to the k-th column
        b = np.divide(np.multiply(s, u), np.sqrt(self._L_documents))
        # f_k corresponds to the k-th column
        f = np.multiply(np.sqrt(self._L_documents), v)

        return PoissonReducedRankParameters(
            logged_dtm=theta,
            alpha=alpha,
            beta=beta,
            f=f,
            b=b,
            svd_v=v,
            svd_u=u,
            svd_s=s,
        )

    @staticmethod
    def wrapper_update_f_poisson_log_likelihood(
        params: np.ndarray, *args
    ) -> np.float32:
        """Wrapper function for the poisson log likelihood, which is used for the scipy optimization routine. It uses f and beta to as parameters and takes the other parameters as given.

        Args:
            params (np.ndarray): parameter vector, which is used for the optimization.
            *args: args is used to process three tuples (alpha, b, y).

        Returns:
            np.float32: poisson log likelihood computed on the given parameters.
        """

        # params contains (beta, f1,f2,...fk)
        alpha, b, y = args
        # b = args[1]

        # construct design matrix
        ones = np.ones(b.shape[0])
        B = np.c_[ones, b]
        # y = args[2]
        logged_mu = alpha + B.dot(params)
        log_likelihood = -1 * poisson_log_likelihood(
            y, np.exp(logged_mu)
        )  # -1 turns maximization into minimization for scipy
        return log_likelihood

    @staticmethod
    def wrapper_update_b_poisson_log_likelihood(
        params: np.ndarray, *args
    ) -> np.float32:
        """Wrapper function used to update the parameter vector b and alpha, taking the other parameters as given.

        Args:
            params (np.ndarray): Parameter vector for the optimization. *args is a tuple of length 3 and used to feed the "given" parameters (beta, f, y).

        Returns:
            np.float32: Computed log likeklihood on the given parameters.
        """

        # params contains (alpha_j, bj1, ... bjK)
        beta, f, y = args

        ones = np.ones(f.shape[1])
        F = np.c_[ones, np.transpose(f)]

        logged_mu = beta + F.dot(params)
        log_likelihood = -1 * poisson_log_likelihood(
            y, np.exp(logged_mu)
        )  # -1 turns maximization into minimization
        return log_likelihood

    def fit_update(self):
        raise NotImplementedError

    def infer_vector(self):
        raise NotImplementedError

    def plot_word_weights(self, params: PoissonReducedRankParameters):
        raise NotImplementedError



from pydantic import BaseModel, field_validator
from typing import List, Optional
from dataclasses import dataclass


class OxfordAPIResponse(BaseModel):
    id: str = None
    definition: str = None
    examples: Optional[List[str]] = None

    class Config:
        allow_population_by_field_name = True

    def __call__(self, **kwargs):
        self.id = kwargs['id']
        self.definition = kwargs['definition']
        self.examples = kwargs['examples']

        return self

    @field_validator('examples')
    def min_len_examples(cls, v):
        if not len(v) >= 1:
            raise ValueError(
                f'Not Enough examples to compile, given: {len(v)}, expected at least 1'
            )
        return v[:10]


@dataclass(init=True, frozen=True, repr=True)
class Words:
    word: str
    senses: List[OxfordAPIResponse]


class WordSimilarities(BaseModel):
    word: str
    year: int
    sense_ids: List[str]
    props: List[float]


class WordFitted(BaseModel):
    word: str
    sense: str
    years: List[int]
    props: List[float]
    poly_fit: List[float]


class SenseEmbedding(BaseModel):
    id: str = None
    definition: str = None
    embedding: List[float] = None

    class Config:
        allow_population_by_field_name = True

    def __call__(self, **kwargs):
        self.id = kwargs['id']
        self.definition = kwargs['definition']
        self.embedding = kwargs['embedding']

        return self


@dataclass(init=True, frozen=True, repr=True)
class WordSenseEmbedding:
    id: str
    definition: str
    embedding: List[float]

    def __str__(self):
        return f'Sense Id: {self.id}\nSense Definition: {self.definition}\nSense Embedding: {len(self.embedding)}'


@dataclass(init=True, frozen=True, repr=True)
class SenseEmbedding:
    word: str
    senses: List[WordSenseEmbedding]


class Embedding(BaseModel):
    word: str = None
    sentence_number_index: List[List] = None
    embeddings: List[List] = None

    class Config:
        allow_population_by_field_name = True

    def __call__(self, **kwargs):
        self.word = kwargs['word']
        self.sentence_number_index = kwargs['sentence_number_index']
        self.embeddings = kwargs['embeddings']

        return self



import os
import shutil
import unittest
import numpy as np
import pandas as pd
import sys
sys.path.append('../src')
from ttta.methods.rolling_lda import RollingLDA
from ttta.methods.lda_prototype import LDAPrototype

class TestRollingLDA(unittest.TestCase):
    def setUp(self):
        self.texts = [["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"]]
        self.dates = [pd.to_datetime("2021-01-01"), pd.to_datetime("2021-01-02"), pd.to_datetime("2021-01-03"),
                      pd.to_datetime("2021-01-04"), pd.to_datetime("2021-01-05"), pd.to_datetime("2021-01-06"),
                      pd.to_datetime("2021-01-07")] * 3
        self.df = pd.DataFrame({"text": self.texts, "date": self.dates})
        self.df.sort_values("date", inplace=True)
        self.roll = RollingLDA(10.0, prototype=5.0, how="1D", warmup=1.0, min_docs_per_chunk=5.0, verbose=2.0,
                          topic_threshold=[3.0, 0.0], initial_epochs=200.0, subsequent_epochs=100.0,
                          memory=2.0, alpha=0.05, gamma=0.05, min_count=1.0, max_assign=True, prototype_measure="cosine")

    def test_init(self):
        roll = RollingLDA(10)
        self.assertEqual(roll._K, 10)
        self.assertEqual(roll._prototype, 10)
        self.assertEqual(roll._threshold, [5, 0.002])
        self.assertEqual(roll._initial_epochs, 100)
        self.assertEqual(roll._subsequent_epochs, 50)
        self.assertEqual(roll._memory, 3)
        self.assertEqual(roll._warmup, 48)
        self.assertEqual(roll._how, "M")
        self.assertEqual(roll._min_docs_per_chunk, 10 * 10)
        self.assertEqual(roll._verbose, 1)
        self.assertEqual(roll._alpha, 0.1)
        self.assertEqual(roll._gamma, 0.1)
        self.assertEqual(roll._min_count, 2)
        self.assertEqual(roll._max_assign, False)
        self.assertEqual(roll._measure, "jaccard")
        self.assertIsInstance(roll.lda, LDAPrototype)
        roll = RollingLDA(10.0, prototype=5.0, how="1D", warmup=1.0, min_docs_per_chunk=5.0, verbose=2.0,
                          topic_threshold=[3.0, 0.0], initial_epochs=200.0, subsequent_epochs=100.0,
                          memory=2.0, alpha=0.05, gamma=0.05, min_count=1.0, max_assign=True, prototype_measure="cosine")

    def test_init_invalid(self):
        with self.assertRaises(ValueError):
            roll = RollingLDA(-1)
        with self.assertRaises(TypeError):
            roll = RollingLDA("a")
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, prototype=-3)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, prototype="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, topic_threshold="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, topic_threshold=["a", 0.1])
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, topic_threshold=[-1, 1.0])
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, topic_threshold=[1, -1.0])
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, topic_threshold=[1, 1.0])
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, initial_epochs="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, initial_epochs=3.5)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, initial_epochs=-1)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, subsequent_epochs=-5)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, subsequent_epochs="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, subsequent_epochs=3.5)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, memory="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, memory=3.5)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, memory=-1)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, warmup=-2)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, warmup="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, warmup=3.5)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, how=1)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, min_docs_per_chunk="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, min_docs_per_chunk=3.5)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, min_docs_per_chunk=-1)
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, verbose="a")
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, alpha="a")
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, alpha=-1)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, alpha=np.array([-1.0] * 10))
        with self.assertRaises(TypeError):
            roll = RollingLDA(10, alpha=np.array(["a"] * 10))
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, alpha=np.array([0.5] * 5))

    def test_fit(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=2, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df, text_column="text", date_column="date", workers=1.0)
        self.assertIsInstance(roll.chunk_indices, pd.DataFrame)
        self.assertEqual(roll.chunk_indices.shape, (7, 4))
        self.assertTrue(roll.lda.is_trained())
        self.assertEqual(roll._last_text["index"], len(self.texts) - 1)
        self.assertEqual(roll._last_text["date"], self.dates[-1])

    def test_fit_invalid(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=2, memory=2, min_docs_per_chunk=1)
        with self.assertRaises(TypeError):
            roll.fit(self.df, text_column=1)
        with self.assertRaises(ValueError):
            roll.fit(self.df, text_column="a")
        with self.assertRaises(TypeError):
            roll.fit(self.df, date_column=1)
        with self.assertRaises(ValueError):
            roll.fit(self.df, date_column="a")
        with self.assertRaises(TypeError):
            roll.fit("a")
        with self.assertRaises(TypeError):
            roll.fit(self.df, text_column="date")
        with self.assertRaises(TypeError):
            roll.fit(self.df, workers=0.5)
        with self.assertRaises(ValueError):
            roll.fit(self.df)
            roll.fit(self.df)
        with self.assertRaises(ValueError):
            roll = RollingLDA(10, min_count=1, min_docs_per_chunk=1, warmup=100)
            roll.fit(self.df)

    def test_get_parameters(self):
        params = self.roll.get_parameters()
        self.assertEqual(params["_K"], 10)
        self.assertEqual(params["_prototype"], 5)
        self.assertEqual(params["_threshold"], [3, 0.0])
        self.assertEqual(params["_initial_epochs"], 200)
        self.assertEqual(params["_subsequent_epochs"], 100)
        self.assertEqual(params["_memory"], 2)
        self.assertEqual(params["_warmup"], 1)
        self.assertEqual(params["_how"], "1D")
        self.assertEqual(params["_min_docs_per_chunk"], 5)
        self.assertEqual(params["_verbose"], 2)
        self.assertEqual(params["_alpha"], 0.05)
        self.assertEqual(params["_gamma"], 0.05)
        self.assertEqual(params["_min_count"], 1)
        self.assertEqual(params["_max_assign"], True)
        self.assertEqual(params["_measure"], "cosine")

    def test_set_parameters(self):
        roll = RollingLDA(10)
        roll.set_parameters({"_K": 5, "_prototype": 3, "_threshold": [2, 0.1], "_initial_epochs": 100, "_subsequent_epochs": 50,
                            "_memory": 3, "_warmup": 2, "_how": "2D", "_min_docs_per_chunk": 10, "_verbose": 1, "_alpha": 0.1,
                            "_gamma": 0.1, "_min_count": 2, "_max_assign": False, "_measure": "jaccard"})
        self.assertEqual(roll._K, 5)
        self.assertEqual(roll._prototype, 3)
        self.assertEqual(roll._threshold, [2, 0.1])
        self.assertEqual(roll._initial_epochs, 100)
        self.assertEqual(roll._subsequent_epochs, 50)
        self.assertEqual(roll._memory, 3)
        self.assertEqual(roll._warmup, 2)
        self.assertEqual(roll._how, "2D")
        self.assertEqual(roll._min_docs_per_chunk, 10)
        self.assertEqual(roll._verbose, 1)
        self.assertEqual(roll._alpha, 0.1)
        self.assertEqual(roll._gamma, 0.1)
        self.assertEqual(roll._min_count, 2)
        self.assertEqual(roll._max_assign, False)
        self.assertEqual(roll._measure, "jaccard")
        with self.assertRaises(TypeError):
            roll.set_parameters(1)

    def test_save_and_load(self):
        self.roll.save("test.pickle")
        roll = RollingLDA(5)
        roll.load("test.pickle")
        self.assertEqual(roll._K, 10)
        self.assertEqual(roll._prototype, 5)
        self.assertEqual(roll._threshold, [3, 0.0])
        self.assertEqual(roll._initial_epochs, 200)
        self.assertEqual(roll._subsequent_epochs, 100)
        self.assertEqual(roll._memory, 2)
        self.assertEqual(roll._warmup, 1)
        self.assertEqual(roll._how, "1D")
        self.assertEqual(roll._min_docs_per_chunk, 5)
        self.assertEqual(roll._verbose, 2)
        self.assertEqual(roll._alpha, 0.05)
        self.assertEqual(roll._gamma, 0.05)
        self.assertEqual(roll._min_count, 1)
        self.assertEqual(roll._max_assign, True)
        self.assertEqual(roll._measure, "cosine")
        os.remove("test.pickle")
        with self.assertRaises(FileNotFoundError):
            roll.load("test.pickle")
        with self.assertRaises(TypeError):
            roll.load(1)
        with self.assertRaises(TypeError):
            roll.save(1)

    def test_fit_update(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df.iloc[:9], text_column="text", date_column="date", workers=1.0)
        roll.fit_update(self.df[9:], text_column="text", date_column="date", workers=1.0)
        self.assertEqual(roll._last_text["index"], len(self.texts) - 1)
        self.assertEqual(roll._last_text["date"], self.dates[-1])
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit_update(self.df, text_column="text", date_column="date", workers=1.0)

    def test_fit_update_invalid(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df[:9], text_column="text", date_column="date", workers=1.0)
        with self.assertRaises(TypeError):
            roll.fit_update("a")
        with self.assertRaises(ValueError):
            roll.fit_update(self.df)
        with self.assertRaises(TypeError):
            roll.fit_update(self.df[9:], text_column=1)
        with self.assertRaises(ValueError):
            roll.fit_update(self.df[9:], text_column="a")
        with self.assertRaises(TypeError):
            roll.fit_update(self.df[9:], date_column=1)
        with self.assertRaises(ValueError):
            roll.fit_update(self.df[9:], date_column="a")
        with self.assertRaises(TypeError):
            roll.fit_update(self.df[9:], text_column="date")
        with self.assertRaises(TypeError):
            roll.fit_update(self.df[9:], workers=0.5)
        with self.assertRaises(TypeError):
            roll.fit_update(self.df[9:], how=0.5)
        with self.assertRaises(TypeError):
            alt_df = self.df.copy()
            alt_df["text"] = alt_df["text"].apply(lambda x: " ".join(x))
            roll.fit_update(self.df[9:], how=0.5)

    def test_top_words(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df, text_column="text", date_column="date", workers=1.0)
        roll.top_words(0.0, number=5.0, topic=1.0, return_as_data_frame=False, )
        self.assertIsInstance(roll.top_words(0, return_as_data_frame=False), list)
        self.assertEqual(len(roll.top_words(0, return_as_data_frame=False)), 10)
        self.assertEqual(len(roll.top_words(0, return_as_data_frame=False)[0]), 5)
        self.assertIsInstance(roll.top_words(return_as_data_frame=False), list)
        self.assertEqual(len(roll.top_words(return_as_data_frame=False)), 10)
        self.assertEqual(len(roll.top_words(return_as_data_frame=False)[0]), 5)
        self.assertIsInstance(roll.top_words(0, return_as_data_frame=True), pd.DataFrame)
        self.assertIsInstance(roll.top_words("all", return_as_data_frame=True), pd.DataFrame)
        self.assertIsInstance(roll.top_words(return_as_data_frame=True), pd.DataFrame)

    def test_top_words_invalid(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df, text_column="text", date_column="date", workers=1.0)
        with self.assertRaises(TypeError):
            roll.top_words(0, return_as_data_frame="a")
        with self.assertRaises(TypeError):
            roll.top_words(0.5)
        with self.assertRaises(ValueError):
            roll.top_words(0, number=0)
        with self.assertRaises(TypeError):
            roll.top_words(0, number=3.5)
        with self.assertRaises(TypeError):
            roll.top_words(0, topic=3.5)
        with self.assertRaises(TypeError):
            roll.top_words("a")
        with self.assertRaises(TypeError):
            roll.top_words(1, importance="a")

    def test_get_word_topic_matrix(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df, text_column="text", date_column="date", workers=1.0)
        with self.assertRaises(TypeError):
            roll.get_word_topic_matrix("a")
        roll.get_word_topic_matrix(0)

    def test_get_time_indices(self):
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how=[pd.to_datetime("2021-01-01"), pd.to_datetime("2021-01-02")], warmup=0, memory=1, min_docs_per_chunk=1)
        roll.fit(self.df[:6], text_column="text", date_column="date", workers=1.0)
        roll._get_time_indices(self.df[6:], how="1D")
        with self.assertRaises(TypeError):
            roll._get_time_indices(1)
        with self.assertRaises(TypeError):
            roll._get_time_indices(self.df[6:], update=1)
        roll = RollingLDA(10, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=0, memory=1, min_docs_per_chunk=1)
        roll.fit(self.df[:6], text_column="text", date_column="date", workers=1.0)
        roll._get_time_indices(self.df[6:], how="2D")
        roll._get_time_indices(self.df[6:], how=self.dates[3:7])

    def test_wordclouds(self):
        roll = RollingLDA(3, min_count=1, topic_threshold=[1, 0.001], prototype=1, how="1D", warmup=1, memory=2, min_docs_per_chunk=1)
        roll.fit(self.df, text_column="text", date_column="date", workers=1.0)
        roll.wordclouds()
        with self.assertRaises(TypeError):
            roll.wordclouds(0)
        with self.assertRaises(TypeError):
            roll.wordclouds(path=1)
        with self.assertRaises(TypeError):
            roll.wordclouds(topic=1.5)
        with self.assertRaises(TypeError):
            roll.wordclouds(number=1.5)
        with self.assertRaises(TypeError):
            roll.wordclouds(height=1.5)
        with self.assertRaises(TypeError):
            roll.wordclouds(width=1.5)
        with self.assertRaises(TypeError):
            roll.wordclouds(show=1.5)
        with self.assertRaises(TypeError):
            roll.wordclouds([0.5])


import pickle
import warnings
from typing import List, Union, Optional, Tuple, Literal
import pandas as pd
from gensim.models import Word2Vec
from ttta.methods.word2vec.word2vec import Word2VecTrainer, Word2VecAlign, Word2VecInference
import numpy as np
from sklearn.manifold import TSNE
import matplotlib.pyplot as plt
from tqdm import tqdm
from ttta.preprocessing.chunk_creation import _get_time_indices
from datetime import datetime

class Word2VecSemanticShift:
    """
    Word2VecSemanticShift class for training and aligning Word2Vec models for semantic shift analysis.

    Methods:
        fit: Fits the Word2Vec models on the texts.
        fit_update: Updates the Word2Vec models on the texts.
        _align_models: Aligns the Word2Vec models.
        save: Saves this class to a file
        load: Loads this class from a file.
        get_parameters: Returns the parameters of the RollingLDA model.
        infer_vector: Infers the vector of a word.
        top_words: Returns the top words similar to a word.
        visualize: Visualizes the semantic shift of a word across chunks
    """
    def __init__(
            self,
            how: Union[str, List[datetime]] = "M",
            min_count: int = 2,
            min_docs_per_chunk: int = None,
            window: int = 5,
            negative: int = 5,
            ns_exponent: float = 0.75,
            vector_size: int = 300,
            alpha: float = 0.025,
            max_vocab_size: Optional[int] = None,
            sample: float = 1e-3, 
            seed: int = 1,
            workers: int = 3,
            min_alpha: float = 0.0001,
            sg: int = 0,
            hs: int = 0,
            cbow_mean: int = 1,
            null_word: int = 0,
            trim_rule: Optional[str] = None,
            sorted_vocab: int = 1,
            compute_loss: bool = False,
            callbacks: Optional[List] = None,
            comment: Optional[str] = None,
            max_final_vocab: Optional[int] = None, 
            shrink_windows: bool = False,
            verbose: int = 1,
            ):
        """
        Args:
        how: List of datetime dates indicating the end of time chunks or a string indicating the frequency of the time chunks as in pandas.resample().
        min_count: Ignores all words with total frequency lower than this.
        min_docs_per_chunk: The minimum number of documents a chunk must contain to be used for the training.
        window: The maximum distance between the current and predicted word within a sentence.
        negative: If > 0, negative sampling will be used, the int for negative specifies how many "noise words" should be drawn. If set to 0, no negative sampling is used.
        ns_exponent: The exponent used to shape the negative sampling distribution. A value of 1.0 samples exactly in proportion to the frequencies, 0.0 samples all words equally, while a negative value samples low-frequency words more than high-frequency words. The popular default value of 0.75 was chosen by the original Word2Vec paper.
        vector_size: Dimensionality of the word vectors.
        alpha: The initial learning rate.
        max_vocab_size: Limits the RAM during vocabulary building. If there are more unique words than this, then prune the infrequent ones. Every 10 million word types need about 1GB of RAM. Set to None for no limit.
        sample: The threshold for configuring which higher-frequency words are randomly downsampled. Highly influencial.
        seed: Seed for the random number generator.
        workers: Use these many worker threads to train the model.
        min_alpha: Learning rate will linearly drop to min_alpha as training progresses.
        sg: Training algorithm: 1 for skip-gram; otherwise CBOW.
        hs: If 1, hierarchical softmax will be used for model training. If set to 0, and negative is non-zero, negative sampling will be used.
        cbow_mean: If 0, use the sum of the context word vectors. If 1, use the mean. Only applies when cbow is used.
        null_word: If > 0, the model uses the null word to fit the model to the training data. If 0, no null word is used.
        trim_rule: Vocabulary trimming rule, specifies whether certain words should remain in the vocabulary, be trimmed away, or handled using the default. Can be None (min_count will be used), or a string that specifies the rule.
        sorted_vocab: If 1, sort the vocabulary by descending frequency before assigning word indexes.
        compute_loss: If True, computes and stores loss value which can be retrieved using get_latest_training_loss().
        callbacks: Sequence of callbacks to be executed at specific stages during training.
        comment: A comment to be added to the model.
        max_final_vocab: Limits the vocabulary size when building the vocabulary. If there are more unique words than this, then prune the infrequent ones. Every 10 million word types need about 1GB of RAM. Set to None for no limit.
        shrink_windows: Whether to shrink the window size as training progresses.
        verbose: The verbosity level. 0 does not print anything, 1 prints the current progress, 2 prints additional debug information.
        """
        self.trainer_args = {
            "min_count":min_count,
            "window":window,
            "negative":negative,
            "ns_exponent":ns_exponent,
            "vector_size":vector_size,
            "alpha":alpha,
            "max_vocab_size":max_vocab_size,
            "sample":sample,
            "seed":seed,
            "workers":workers,
            "min_alpha":min_alpha,
            "sg":sg,
            "hs":hs,
            "cbow_mean":cbow_mean,
            "null_word":null_word,
            "trim_rule":trim_rule,
            "sorted_vocab":sorted_vocab,
            "compute_loss":compute_loss,
            "callbacks":callbacks,
            "comment":comment,
            "max_final_vocab":max_final_vocab,
            "shrink_windows":shrink_windows
        }
        self.reference = None
        self.how = how
        self.trainers: List[Word2VecTrainer] = []
        self.word2vecs: List[Word2Vec] = []
        self.chunks: List[str] = []
        self.aligned_models: List[Word2Vec] = []
        self._verbose = verbose
        self.chunk_indices = None
        self.min_docs_per_chunk = min_docs_per_chunk

    def is_trained(self) -> bool:
        """
        Returns whether the model has been trained or not.
        Returns:
            A boolean value indicating whether the model has been trained or not.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> ss.is_trained()
            False
        """
        return len(self.word2vecs) > 0
    
    def get_chunks(self) -> List[str]:
        """
        Returns the chunks of the model.
        Returns:
            A list of the chunks of the model.
        """
        return self.chunks
    
    def get_reference(self) -> int:
        """
        Returns the reference chunk index.
        Returns:
            The reference chunk index.
        """
        return self.reference

    def fit(
            self,
            texts: pd.DataFrame,
            text_column: str = "text",
            date_column: str = "date",
            date_format: str = "%Y-%m-%d",
            align_reference: int = -1,
            epochs: int = 5,
            start_alpha: float = 0.025,
            end_alpha: float = 0.0001,
        ) -> None:
        """
        Fits the Word2Vec models on the texts.
        Args:
            texts: The texts to train the Word2Vec models on.
            text_column: The column containing the texts.
            date_column: The column containing the dates.
            date_format: The format of the dates.
            align_reference: The reference chunk index to align the models.
            epochs: The number of epochs to train the models.
            start_alpha: The initial learning rate.
            end_alpha: The final learning rate.

        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> data = pd.read_csv("data.csv")
            >>> ss.fit(data, text_column="text", date_column="date", date_format="%Y-%m-%d", align_reference=-1, epochs=5, start_alpha=0.025, end_alpha=0.0001, date_groupby="year")
        """

        if not isinstance(texts, pd.DataFrame):
            raise TypeError("texts must be a pandas DataFrame!")
        if not isinstance(text_column, str):
            raise TypeError("text_column must be a string!")
        if text_column not in texts.columns:
            raise ValueError("texts must contain the column specified in text_column!")
        if not isinstance(date_column, str):
            raise TypeError("date_column must be a string!")
        if date_column not in texts.columns:
            raise ValueError("texts must contain the column specified in date_column!")
        
        self._text_column = text_column
        self._date_column = date_column


        if isinstance(texts[self._text_column].iloc[0], list):
            if not isinstance(texts[self._text_column].iloc[0][0], str):
                raise TypeError("text column must be contain list of strings or lists of lists of strings!")
        elif isinstance(texts[self._text_column].iloc[0], str):
            texts[self._text_column] = texts[self._text_column].apply(lambda x: x.split())
            warnings.warn("The elements of the text column contain a string. "
                          "Beware that the texts might not be preprocessed")
        else:
            raise TypeError(
                "text column must be contain list of strings or lists of lists of strings!")
        
        if self.is_trained():
            raise ValueError("The model has already been trained. Call 'fit_update' to update the model.")

        texts[self._date_column] = pd.to_datetime(texts[self._date_column], format=date_format)
        texts.sort_values(by=self._date_column, inplace=True)
        self.chunk_indices = _get_time_indices(texts, how=self.how, date_column=self._date_column, min_docs_per_chunk=self.min_docs_per_chunk)

        iterator = self.chunk_indices.iterrows()
        if self._verbose > 0:
            iterator = tqdm(iterator, total=len(self.chunk_indices), unit="chunk")
        for i, row in enumerate(iterator):
            date = self.chunk_indices["Date"].iloc[i]
            trainer = Word2VecTrainer(**self.trainer_args)
            if self._verbose > 0 and i < len(self.chunk_indices) - 1:
                iterator.set_description(f"Chunk {i + 1}/{len(self.chunk_indices)}")
            if i == 0:  # fit warmup chunks
                trainer.train(texts[self._text_column].iloc[:self.chunk_indices["chunk_start"].iloc[1]],
                              epochs=epochs, start_alpha=start_alpha, end_alpha=end_alpha)
                continue
            end = len(texts) if i + 1 >= len(self.chunk_indices) else int(self.chunk_indices.iloc[i + 1]["chunk_start"] - 1)
            trainer.train(texts[text_column].iloc[self.chunk_indices["chunk_start"].iloc[i]:end], epochs=epochs, start_alpha=start_alpha, end_alpha=end_alpha)

            self.trainers.append(trainer)
            self.word2vecs.append(trainer.get_model())
            self.chunks.append(str(date))

        
        self.reference = align_reference
        self._align_models(self.word2vecs, reference=align_reference)

    def fit_update(
            self,
            texts: pd.DataFrame,
            text_column: str = "text",
            date_column: str = "date",
            align_reference: int = -1,
            epochs: int = 5,
            start_alpha: float = 0.025,
            end_alpha: float = 0.0001
        ) -> None:
        """
        Updates the Word2Vec models on the texts.
        Args:
            texts: The texts to train the Word2Vec models on.
            text_column: The column containing the texts.
            date_column: The column containing the dates.
            align_reference: The reference chunk index to align the models.
            epochs: The number of epochs to train the models.
            start_alpha: The initial learning rate.
            end_alpha: The final learning rate.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> data = pd.read_csv("data.csv")
            >>> ss.fit_update(data, text_column="text", date_column="date", align_reference=-1, epochs=5, start_alpha=0.025, end_alpha=0.0001)
        """

        if self.word2vecs is None:
            raise ValueError("Initial training has not been done. Call fit first.")
        
        if not isinstance(texts, pd.DataFrame):
            raise TypeError("texts must be a pandas DataFrame!")
        if not isinstance(text_column, str):
            raise TypeError("text_column must be a string!")
        if text_column not in texts.columns:
            raise ValueError("texts must contain the column specified in text_column!")
        if not isinstance(date_column, str):
            raise TypeError("date_column must be a string!")
        if date_column not in texts.columns:
            raise ValueError("texts must contain the column specified in date_column!")
        self._text_column = text_column
        self._date_column = date_column

        if not isinstance(texts[self._text_column].iloc[0], str):
            raise TypeError("The elements of the 'texts' column of texts must each contain a string!")
        
        if not self.is_trained():
            raise ValueError("The model has not been trained. Call 'fit' to train the model first.")

        texts[date_column] = pd.to_datetime(texts[date_column])
        texts.sort_values(by=date_column, inplace=True)

        grouped_texts = texts.groupby(date_column)
        for date, group in grouped_texts:
            try:
                idx = self.chunks.index(str(date))
            except ValueError:
                raise ValueError(f"The date: {str(date)}, is not in the training chunks. Call fit instead.")
            # print(f"Training on chunk: {date}")
            self.trainers[idx].train(group[text_column].tolist(), epochs=epochs, start_alpha=start_alpha, end_alpha=end_alpha, update=True)
            self.word2vecs[idx] = self.trainers[idx].get_model()


        self.reference = align_reference
        self._align_models(self.word2vecs, reference=align_reference)

    def _align_models(self, models: List[Word2Vec], reference: int = -1) -> None:
        """
        Aligns the Word2Vec models.
        Args:
            models: The Word2Vec models to align.
            reference: The reference chunk index.
        """
        self.aligner = Word2VecAlign(models)
        self.aligned_models = self.aligner.align(reference=reference)

    def save(self, path: str) -> None:
        """
            Saves the Word2Vec Semantic Shift model to the given path as a .pickle-file.
            Args:
                path: The path to which the Word2Vec Semantic Shift model should be saved.
            Returns:
                None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        pickle.dump(self, open(path, "wb"))

    def load(self, path: str) -> None:
        """
            Loads a pickled Word2Vec Semantic Shift model from the given path.
            Args:
                path: The path from which the Word2Vec Semantic Shift model should be loaded.
            Returns:
                None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        loaded = pickle.load(open(path, "rb"))
        self.__dict__ = loaded.__dict__

    def get_parameters(self) -> dict:
        """
        Returns the parameters of the Word2Vec models.
        Returns:
            A dictionary containing the parameters of the Word2Vec models.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> ss.get_parameters()
            {'min_count': 2, 'window': 5, 'negative': 5, 'ns_exponent': 0.75, 'vector_size': 300, 'alpha': 0.025, 'max_vocab_size': None, 'sample': 0.001, 'seed': 1, 'workers': 3, 'min_alpha': 0.0001, 'sg': 0, 'hs': 0, 'cbow_mean': 1, 'null_word': 0, 'trim_rule': None, 'sorted_vocab': 1, 'compute_loss': False, 'callbacks': None, 'comment': None, 'max_final_vocab': None, 'shrink_windows': False}
        
        """
        return self.__dict__.copy()

    def infer_vector(self, word: str, chunk_index: int, norm: bool = False, aligned: bool = True) -> np.ndarray:
        """
        Infers the vector of a word.
        Args:
            word: The word to infer the vector.
            chunk_index: The index of the chunk.
            norm: Whether to normalize the vector or not.
            aligned: Whether to infer the vector from the aligned models or not.
        Returns:
            The inferred vector of the word.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> ss.infer_vector("trump", 0, aligned=True)
            array([ 0.001,  0.002,  0.003, ..., -0.001,  0.002,  0.001], dtype = float32)
        """
        if aligned:
            if not self.aligned_models:
                raise ValueError("Models are not aligned. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.aligned_models[chunk_index])
            return inferencer.infer_vector(word, norm=norm)
        
        else:
            if not self.word2vecs:
                raise ValueError("Models are not trained. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.word2vecs[chunk_index])
            return inferencer.infer_vector(word, norm=norm)

    def top_words(self, word: str, chuck_index: int, k: int = 10, pos_tag: Union[bool, str, List[str]] = False, aligned: bool = True) -> Tuple[List[str], List[float]]:
        """
        Returns the top words similar to a word.
        Args:
            word: The word to get the top words.
            chuck_index: The index of the chunk.
            k: The number of top words to return.
            pos_tag: The part-of-speech tag of the words to return.
            aligned: Whether to get the top words from the aligned models or not.
        Returns:
            A tuple containing the top words and their similarities.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> ss.top_words("trump", 0, k=2, pos_tag=False, aligned=True)
            (['donald', 'president'], [0.8, 0.7])
        """
        if aligned:
            if not self.aligned_models:
                raise ValueError("Models are not aligned. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.aligned_models[chuck_index])
            top_words, sims = inferencer.get_top_k_words(word, k=k, pos_tag=pos_tag)
            return top_words, sims
    
        else:
            if not self.word2vecs:
                raise ValueError("Models are not trained. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.word2vecs[chuck_index])
            top_words, sims = inferencer.get_top_k_words(word, k=k, pos_tag=pos_tag)
            return top_words, sims
       
    def get_vocab(self, chunk: Union[int,str] = "reference") -> List[str]:
        """
        Returns the vocabulary of a chunk.
        Args:
            chunk: The index of the chunk or 'reference'.
        Returns:
            The vocabulary of the chunk.
        """
        if chunk == "reference":
            return self.word2vecs[-1].wv.index_to_key
        elif isinstance(chunk, int):
            return self.word2vecs[chunk].wv.index_to_key
        else:
            raise ValueError("chunk must be an integer or 'reference'.")

    def visualize(
            self, 
            main_word: str, 
            chunks_tocompare: Optional[List[int]] = None,
            reference: Optional[int] = -1,
            k: int = 10, 
            pos_tag: Union[bool, str, List[str]] = False, 
            extra_words: Optional[List[str]] = None, 
            ignore_words: Optional[List[str]] = None,
            aligned: bool = True,
            tsne_perplexity: int = 30,
            tsne_metric: str = 'euclidean',
            tsne_learning_rate: Union[str, int] = 'auto'
            ) -> None:
        """
        Visualizes the semantic shift of a word across chunks.
        Args:
            main_word: The main word to visualize.
            chunks_tocompare: The chunks to compare. If None, all chunks will be compared.
            reference: The reference chunk index.
            k: The number of top words to return.
            pos_tag: The part-of-speech tag of the words to return.
            extra_words: The extra words to include in the visualization.
            ignore_words: The words to ignore in the visualization.
            aligned: Whether to visualize the aligned models or not.
            tsne_perplexity: The perplexity of the t-SNE algorithm.
            tsne_metric: The metric of the t-SNE algorithm.
            tsne_learning_rate: The learning rate of the t-SNE algorithm.
        
        Examples:
            >>> ss = Word2VecSemanticShift()
            >>> ss.visualize(
            ...     reference=-1,
            ...     chunks_tocompare=['1980', '2000', '2015', '2017'],
            ...     main_word="trump",
            ...     k=30,
            ...     pos_tag=False,
            ...     extra_words= [
            ...         "game",
            ...         "play",
            ...         "king",
            ...         "jack",
            ...         "heart"
            ...     ],
            ...     ignore_words= [
            ...         "use",
            ...         "give",
            ...         "would"
            ...     ],
            ...     aligned=True,
            ...     tsne_perplexity=10,
            ...     tsne_metric='euclidean',
            ...     tsne_learning_rate='auto'
            ... )
        """
        if aligned:
            if len(self.aligned_models) == 0:
                raise ValueError("Models are not aligned. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.aligned_models[reference])
        else:
            if len(self.word2vecs) == 0:
                raise ValueError("Models are not trained. Call fit or fit_update first.")

            inferencer = Word2VecInference(self.word2vecs[reference])

        plot_vocab = extra_words if extra_words is not None else []

        try:
            words, _ = inferencer.get_top_k_words(main_word, k=k, pos_tag=pos_tag)
        
        except ValueError:
            raise ValueError(f"Word: {main_word} is not in the vocabulary of the model.")
        
        
        plot_vocab.extend(words)
        plot_vocab = list(set(plot_vocab))

        if ignore_words is not None:
            for word in ignore_words:
                if word in plot_vocab:
                    plot_vocab.remove(word)

        embeddings = []
        not_in_vocab = []

        for word in plot_vocab:
            if word != main_word:
                try:
                    vector = inferencer.infer_vector(word)
                    embeddings.append(vector)
                
                except ValueError:
                    not_in_vocab.append(word)

        for word in not_in_vocab:
            plot_vocab.remove(word)

        if chunks_tocompare is None:
            chunks_tocompare = list(range(len(self.chunks)))
        
        else:
            for chunk in chunks_tocompare:
                if chunk not in self.chunks:
                    raise ValueError(f"Chunk: {chunk} is not in the training chunks. Choose from: {self.chunks}")

            chunks_tocompare = [self.chunks.index(chunk) for chunk in chunks_tocompare]

        main_word_embeddings = []
        # for chuck in range(len(self.chunks)):
        for chuck_idx in chunks_tocompare:
            inferencer = Word2VecInference(self.aligned_models[chuck_idx])
            try:
                vector = inferencer.infer_vector(main_word)
                main_word_embeddings.append(vector)
                plot_vocab.append(f'{main_word}_{self.chunks[chuck_idx]}')

            except ValueError:
                raise ValueError(f"Word: {main_word} is not in the vocabulary of the model in chunk: {self.chunks[chuck_idx]}")
        
        X = np.array(embeddings + main_word_embeddings)
        n_samples = X.shape[0]
        if n_samples < 2:
            raise ValueError("Number of samples must be greater than 1.")
        
        if tsne_perplexity > n_samples - 1:
            raise ValueError(f"Perplexity must be less than the number of samples - 1= {n_samples - 1}")
        

        X_embedded = TSNE(
            metric= tsne_metric,
            learning_rate= tsne_learning_rate,
            perplexity= tsne_perplexity
            ).fit_transform(X)

        df = pd.DataFrame(X_embedded)
        df['word'] = plot_vocab
        df_words = df.iloc[:-len(chunks_tocompare)]
        df_main_word = df.iloc[-len(chunks_tocompare):]

        # Plotting
        fig, ax = plt.subplots(1, figsize=(14,10), dpi=80, facecolor='#ffffff')
        ax.set_facecolor('#ffffff')

        # Plot the words' embeddings
        x_words = df_words[0]
        y_words = df_words[1]
        plt.scatter(x_words, y_words, c="0.6")

        for i in range(len(df_words)):
            ax.annotate(df_words['word'].iloc[i], (x_words.iloc[i], y_words.iloc[i]), color='#71797E', fontsize=14)

        # Plot the main_word's embeddings across all chunks
        x_main_word = df_main_word[0]
        y_main_word = df_main_word[1]
        plt.scatter(x_main_word, y_main_word, c="red")

        for i in range(len(df_main_word)):
            ax.annotate(df_main_word['word'].iloc[i], (x_main_word.iloc[i], y_main_word.iloc[i]), color='black', fontsize=17)

        # Draw arrows between main_word embeddings across chunks

        def drawArrow(A, B):
            plt.arrow(A[0], A[1], B[0] - A[0], B[1] - A[1], head_width=0.01, length_includes_head=True, color='black')
        
        for i in range(1, len(x_main_word)):
            drawArrow([x_main_word.iloc[i-1], y_main_word.iloc[i-1]], [x_main_word.iloc[i], y_main_word.iloc[i]])

        # Show the plot
        plt.savefig(f'{main_word}.png')
        plt.show()



import os
import logging
import tqdm
from typing import Union, List, Optional
from pathlib import Path
import math
from transformers import RobertaTokenizer, RobertaForMaskedLM, RobertaModel, DataCollatorForLanguageModeling, get_linear_schedule_with_warmup, logging as lg
import torch
from torch.utils.data import DataLoader, Dataset
import torch.optim as optim
from accelerate import Accelerator
from nltk.corpus import stopwords
from nltk.tag import pos_tag
import collections
import random
# import nltk
# nltk.download('averaged_perceptron_tagger')


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class CustomDataset(Dataset):
    """
    This class is used to create a custom dataset for the Roberta model. It inherits from torch.utils.data.Dataset.
    
    Methods
    -------
        __init__(data: List[str], tokenizer, max_length=128, truncation=True, padding=True)
            The constructor for the CustomDataset class.
        __len__()
            This method is used to get the length of the dataset.
        __getitem__(idx)
            This method is used to get the item at a specific index.  
    """
    def __init__(
            self, 
            data: List[str], 
            tokenizer, 
            max_length=128,
            truncation=True,
            padding= "max_length",
            ):
        """
        Args:
            data (List[str]): List of strings to create a dataset from.
            tokenizer: Tokenizer to tokenize the data with.
            max_length (int): Maximum length of the input sequence. Defaults to 128.
            truncation (bool): Whether to truncate the input sequence to max_length or not. Defaults to True.
            padding (str): Whether to pad the input sequence to max_length or not. Defaults to "max_length".
        
        Attributes:
            tokenizer: Tokenizer to tokenize the data with.
            max_length (int): Maximum length of the input sequence. Defaults to 128.
            tokenized_data (dict): Dictionary containing the input_ids, attention_mask, and labels. 
        """
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.tokenized_data = tokenizer(data, truncation=truncation, padding=padding, max_length=max_length)

    def __len__(self):
        return len(self.tokenized_data.input_ids)

    def __getitem__(self, idx):
        """
        Retrieves the item at the specified index.

        Parameters:
            idx (int): Index of the item to retrieve.

        Returns:
            tokenized_data (dict): Dictionary containing the input_ids, attention_mask, and labels.
        """
        # Get the tokenized inputs at the specified index
        input_ids = self.tokenized_data.input_ids[idx]
        attention_mask = self.tokenized_data.attention_mask[idx]

        # Return a dictionary containing input_ids, attention_mask, and labels (if applicable)
        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask
            # Add 'labels': labels if you have labels for your data
        }

class RobertaTrainer:
    """
    This class is used to train a Roberta model.

    Methods
    -------
        __init__(model_name="roberta-base", max_length=128, mlm_probability=0.15, batch_size=4, learning_rate=1e-5, epochs=3, warmup_steps=500, split_ratio=0.8)
            The constructor for the RobertaTrainer class.
        prepare_dataset(data)
            This method is used to prepare the dataset for training.
        train(data, output_dir: Union[str, Path] = None)
            This method is used to train the model.
    """
    def __init__(
            self, 
            model_name: str = "roberta-base", 
            max_length: int = 128, 
            mlm_probability: float = 0.15, 
            batch_size: int = 4, 
            learning_rate: float = 1e-5, 
            epochs: int = 3, 
            warmup_steps: int = 500, 
            split_ratio: float = 0.8, 
            truncation: bool = True, 
            padding: str = "max_length"
            ):

        """
        Args:
            model_name (str): Name of the model to train. Defaults to "roberta-base".
            max_length (int): Maximum length of the input sequence. Defaults to 128.
            mlm_probability (float): Probability of masking tokens in the input sequence. Defaults to 0.15.
            batch_size (int): Size of the batch. Defaults to 4.
            learning_rate (float): Learning rate of the optimizer. Defaults to 1e-5.
            epochs (int): Number of epochs to train the model for. Defaults to 3.
            warmup_steps (int): Number of warmup steps for the learning rate scheduler. Defaults to 500.
            split_ratio (float): Ratio to split the data into train and test. Defaults to 0.8.
            truncation (bool): Whether to truncate the input sequence to max_length or not. Defaults to True.
            padding (str): Whether to pad the input sequence to max_length or not. Defaults to "max_length".
        
        Attributes:
            tokenizer (transformers.RobertaTokenizer): Tokenizer to tokenize the data with.
            model (transformers.RobertaForMaskedLM): Model to train.
            data_collator (transformers.DataCollatorForLanguageModeling): DataCollatorForLanguageModeling object to collate the data.
            split_ratio (float): Ratio to split the data into train and test. Defaults to 0.8.
            truncation (bool): Whether to truncate the input sequence to max_length or not. Defaults to True.
            padding (str): Whether to pad the input sequence to max_length or not. Defaults to "max_length".
            max_length (int): Maximum length of the input sequence. Defaults to 128.
            batch_size (int): Size of the batch. Defaults to 4.
            learning_rate (float): Learning rate of the optimizer. Defaults to 1e-5.
            epochs (int): Number of epochs to train the model for. Defaults to 3.
            warmup_steps (int): Number of warmup steps for the learning rate scheduler. Defaults to 500.
            accelerator (accelerate.Accelerator): Accelerator object to distribute the training across multiple GPUs.
        """
        
        
        self.tokenizer = RobertaTokenizer.from_pretrained(model_name)
        self.model = RobertaForMaskedLM.from_pretrained(model_name)

        self.data_collator = DataCollatorForLanguageModeling(
            tokenizer=self.tokenizer, 
            mlm=True, 
            mlm_probability=mlm_probability
            )

        self.split_ratio = split_ratio
        self.truncation = truncation
        self.padding = padding
        self.max_length = max_length
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.warmup_steps = warmup_steps
        self.accelerator = Accelerator()

    def prepare_dataset(self, data: List[str]):
        """
        This method is used to prepare the dataset for training.
        Args:
            data: List of strings to train the model on.
            
        Returns:
            train_loader (torch.utils.data.DataLoader): DataLoader object containing the training data.
            dataset (CustomDataset): CustomDataset object containing the training data.
        """
        dataset = CustomDataset(
            data, 
            self.tokenizer, 
            max_length=self.max_length,
            truncation=self.truncation,
            padding=self.padding
            )
        
        train_loader = DataLoader(
            dataset, 
            batch_size=self.batch_size, 
            shuffle=True, 
            collate_fn=self.data_collator
            )
        
        return train_loader, dataset

    def train(
            self, 
            data: List[str],
            output_dir: Optional[Union[str, Path]] = None
            ) -> None:
        """
        This method is used to train the model.

        Args:
            data (List[str]): List of strings to train the model on.
            output_dir (str, Path, None): Path to save the model to. Defaults to None.
        

        Examples:
            >>> model = RobertaTrainer(epoch=3)
            >>> model.train(data=["The brown fox jumps over the lazy dog", "The brown fox jumps over the lazy dog", "Hello world!"], output_dir="../../output/MLM_roberta")
            Epoch: 0 | Loss: 1.1637206077575684 | Perplexity: 3.2020153999328613
            Epoch: 1 | Loss: 0.6941609382629395 | Perplexity: 2.0011680126190186
            Epoch: 2 | Loss: 0.4749067425727844 | Perplexity: 1.608262062072754  
        """
        
        train_data, test_data = train_test_split(
            data, 
            test_ratio= 1 - self.split_ratio, 
            random_seed=42
            )
        
        train_loader, _ = self.prepare_dataset(train_data)
        test_loader, _ = self.prepare_dataset(test_data)
        
        optimizer = optim.AdamW(
            self.model.parameters(), 
            lr=self.learning_rate
            )

        model, optimizer, train_loader, test_loader = self.accelerator.prepare(
            self.model, 
            optimizer, 
            train_loader, 
            test_loader
            )
        
        scheduler = get_linear_schedule_with_warmup(
            optimizer, 
            num_warmup_steps=self.warmup_steps, 
            num_training_steps=len(train_loader) * self.epochs
            )

        progress_bar = tqdm.tqdm(
            range(len(train_loader) * self.epochs), 
            desc="Training", 
            dynamic_ncols=True
            )
        
        for epoch in range(self.epochs):
            self.model.train()

            for batch in train_loader:
                outputs = self.model(**batch)
                loss = outputs.loss
                self.accelerator.backward(loss)
                optimizer.step()
                scheduler.step()  # Update learning rate scheduler
                optimizer.zero_grad()
                progress_bar.update(1)

            self.model.eval()
            losses = []
            for step, batch in enumerate(test_loader):
                with torch.no_grad():
                    outputs = self.model(**batch)
                
                loss = outputs.loss
                losses.append(self.accelerator.gather(loss.repeat(self.batch_size)))
            
            losses = torch.cat(losses)
            losses = losses[:len(test_data)]

            try:
                perplexity = math.exp(torch.mean(losses))
            except OverflowError:
                perplexity = float("inf")
            print(f"Epoch: {epoch} | Loss: {torch.mean(losses)} | Perplexity: {perplexity}")

            # Save model
            if output_dir is not None:
                self.accelerator.wait_for_everyone()
                unwrapped_model = self.accelerator.unwrap_model(model)
                unwrapped_model.save_pretrained(output_dir, save_function=self.accelerator.save)
                if self.accelerator.is_main_process:
                    self.tokenizer.save_pretrained(output_dir)






class RobertaEmbedding:
    """
    This class is used to infer vector embeddings from a document.

    Methods
    -------
        __init__(pretrained_model_path:Union[str, Path] = None)
            The constructor for the VectorEmbeddings class.
        _roberta_case_preparation()
            This method is used to prepare the Roberta model for the inference.
        infer_vector(doc:str, main_word:str)
            This method is used to infer the vector embeddings of a word from a document.
        infer_mask_logits(doc:str)
            This method is used to infer the logits of a word from a document.
    """
    def __init__(
        self,
        pretrained_model_path: Union[str, Path],
    ):
        """
        Args:
            pretrained_model_path (str, Path, None): Path to the pretrained model. Defaults to None.

        Attributes:
            model_path (str, Path, None): Path to the pretrained model. Defaults to None.
            model (transformers.RobertaModel): RobertaModel object to infer vector embeddings from.
            MLM (transformers.RobertaForMaskedLM): RobertaForMaskedLM object to infer vector embeddings from.
            tokenizer (transformers.RobertaTokenizer): Tokenizer to tokenize the data with.
            max_length (int): Maximum length of the input sequence. Defaults to 128.
            vocab (bool): Whether the model has been initialized or not.
        """
        self.model_path = pretrained_model_path
        self._tokens = []
        self.model = None
        self.vocab = False

        lg.set_verbosity_error()
        self._roberta_case_preparation()

    @property
    def tokens(self):
        return self._tokens

    def _roberta_case_preparation(self) -> None:
        """
        This method is used to prepare the BERT model for the inference.
        """
        model_path = self.model_path if os.path.exists(self.model_path) else 'roberta-base'
        self.tokenizer = RobertaTokenizer.from_pretrained(model_path)
        self.model = RobertaModel.from_pretrained(
            model_path, 
            output_hidden_states=True
            )
        self.MLM = RobertaForMaskedLM.from_pretrained(
            model_path
        )
        # self.model_max_length = self.model.config.max_position_embeddings
        # self.mlm_max_length = self.MLM.config.max_position_embeddings
        self.model.eval()
        self.MLM.eval()
        self.vocab = True

    def infer_vector(self, doc:str, main_word:str) -> Optional[torch.Tensor]:
        """
        This method is used to infer the vector embeddings of a word from a document.

        Args:
            doc (str): Document to process
            main_word (str): Main work to extract the vector embeddings for.

        Returns: 
            embeddings (torch.Tensor): Tensor of stacked embeddings of shape (num_embeddings, embedding_size) where num_embeddings is the number of times the main_word appears in the doc.

        Examples:
            >>> model = RobertaEmbedding()
            >>> model.infer_vector(doc="The brown fox jumps over the lazy dog", main_word="fox")
            tensor([[-0.2182, ..., -0.1709],
                    ...,
                    [-0.2182, ..., -0.1706]])
        """
        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.model.__class__.__name__} has not been initialized'
            )
        
     
        input_ids = self.tokenizer(doc, return_tensors="pt", max_length=512, truncation=True).input_ids
        token = self.tokenizer.encode(main_word, add_special_tokens=False)[0]
        word_token_index = torch.where(input_ids == token)[1]

        try:
            with torch.no_grad():
                embeddings = self.model(input_ids).last_hidden_state
               
            emb = [embeddings[0, idx] for idx in word_token_index]
            return torch.stack(emb).mean(dim=0)
        
        except IndexError:
            raise ValueError(f'The word: "{main_word}" does not exist in the list of tokens')



    
    def infer_mask_logits(self, doc:str) -> Optional[torch.Tensor]:
        """
        This method is used to infer the logits of the mask token in a document.

        Args:
            doc (str): Document to process where the mask token is present.

        Returns: 
            logits (Optional[torch.Tensor]): Tensor of stacked logits of shape (num_embeddings, logits_size) where num_embeddings is the number of times the mask token appears in the doc withing the max_length.

        Examples:
            >>> model = RobertaEmbedding()
            >>> model.infer_mask_logits(doc="The brown fox <mask> over the lazy dog")
            tensor([[-2.1816e-01,  ..., -1.7064e-01],
                    ...,
                    [-2.1816e-01, ..., -1.7093e-01]])
        """

        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.MLM.__class__.__name__} has not been initialized'
            )

        input_ids = self.tokenizer(doc, return_tensors="pt", max_length= 512, truncation=True).input_ids
        mask_token_index = torch.where(input_ids == self.tokenizer.mask_token_id)[1]
        l = []
        try:
            with torch.no_grad():
                logits = self.MLM(input_ids).logits
                
            l = [logits[0, idx] for idx in mask_token_index]
            return torch.stack(l) if len(l) > 0 else torch.empty(0)
        
        except IndexError:
            raise ValueError(f'The mask falls outside of the max length of {512}, please use a smaller document')

        




class RobertaInference:
    """
    Wrapper class for the RobertaEmbedding class for inference.

    Methods
    -------
        __init__(pretrained_model_path:Union[str, Path] = None)
            The constructor for the VectorEmbeddings class.
        _roberta_case_preparation()
            This method is used to prepare the Roberta model for the inference.
        get_embedding(word:str, doc: Optional[Union[str, List[str]]] = None,, mask:bool=False)
            This method is used to infer the vector embeddings of a word from a document.
        get_top_k_words(word:str, doc:str, k:int=3)
            This method is used to infer the vector embeddings of a word from a document.
    """

    def __init__(
            self,
            pretrained_model_path: Optional[Union[str, Path]] = None,
    ):
        """
        Args:
            pretrained_model_path (str, Path, None): Path to the pretrained model. Defaults to None.
        
        Attributes:
            model_path (str, Path, None): Path to the pretrained model. Defaults to None.
            word_vectorizor (RobertaEmbedding): RobertaEmbedding object to infer vector embeddings from.
            vocab (bool): Whether the model has been initialized or not.
        """
        self.model_path = pretrained_model_path
        if pretrained_model_path is not None:
            if not os.path.exists(pretrained_model_path):
                raise ValueError(
                    f'The path {pretrained_model_path} does not exist'
                )
            self.model_path = Path(pretrained_model_path)

        self.word_vectorizor = None
        self.vocab = False
        

        lg.set_verbosity_error()
        self._roberta_case_preparation()
    
    
    def _roberta_case_preparation(self) -> None:
        """
        This method is used to prepare the Roberta model for the inference.
        """
        model_path = self.model_path if self.model_path is not None else 'roberta-base'
        self.tokenizer = RobertaTokenizer.from_pretrained(model_path)
        self.word_vectorizor = RobertaEmbedding(pretrained_model_path=model_path)
        self.vocab = True

    
    def get_embedding(
            self,
            main_word : str, 
            doc: Optional[str] = None,
            mask : bool = False
            ) -> torch.Tensor:
        
        """
        This method is used to infer the vector embeddings of a word from a document.

        Args:
            main_word (str): Word to get the vector embeddings for
            doc (str, None): Documents to get the vector embeddings of the main_word from. If None, the document is the main_word itself. Defaults to None.
            mask: Whether to mask the main_word in the documents or not. Defaults to False.
            
        Returns: 
            embeddings (torch.Tensor): Tensor of stacked embeddings of shape (num_embeddings, embedding_size) where num_embeddings is the number of times the main_word appears in the doc, depending on the mask parameter.

        Examples:
            >>> model = RobertaInference()
            >>> model.get_embedding(main_word="office", doc="The brown office is very big", mask=False)
            tensor([[-0.2182, ..., -0.1709],
                    ...,
                    [-0.2182, ..., -0.1706]])
        """
        
        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.word_vectorizor.__class__.__name__} has not been initialized'
            )
        
        doc = self._cutDoc(doc=doc, main_word=main_word)
        if doc is None:
            emb = torch.empty(0)
        
        
        if mask:
            doc = doc.replace(main_word, self.tokenizer.mask_token)
            main_word = self.tokenizer.mask_token
        
        else:
            main_word = ' ' + main_word.strip()
            
        try:
            emb = self.word_vectorizor.infer_vector(doc=doc, main_word=main_word)
            return emb
        
        except ValueError:
            # print(f'The word: "{main_word}" does not exist in the list of tokens')
            return torch.empty(0)

    def get_top_k_words(
            self,
            main_word : str,
            doc: str,
            k: int = 3,
            pot_tag: Union[bool, str, List[str]] = False
            ) -> List[str]:
        """
        This method is used to infer the vector embeddings of a main_word from a document.
        Args:
            main_word: Word to mask
            doc: Document to infer the top k words of the main_word from
            k: Number of top words to return

        Returns:
            top_k_words (List[str]): List of top k words

        Examples:
            >>> model = RobertaInference()
            >>> model.get_top_k_words(main_word="office", doc="The brown office is very big")
            ['room', 'eye', 'bear']
        """
        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.word_vectorizor.__class__.__name__} has not been initialized'
            )
        
        doc = self._cutDoc(doc=doc, main_word=main_word)
        if doc is None:
            return []

        masked_doc = doc.replace(main_word, '<mask>')
        try:
            logits = self.word_vectorizor.infer_mask_logits(doc=masked_doc)
            top_k = []

            for logit_set in logits:
                top_k_tokens = torch.topk(logit_set, k).indices
                top_k_words = [self.tokenizer.decode(token.item()).strip() for token in top_k_tokens]
                if isinstance(pot_tag, str):
                    pot_tag = [pot_tag]
                
                if isinstance(pot_tag, list):
                    top_k_words = [word for word in top_k_words if pos_tag([word])[0][1] in pot_tag]
                    
                elif pot_tag:
                    main_pos = pos_tag([main_word])[0][1]
                    top_k_words = [word for word in top_k_words if pos_tag([word])[0][1] == main_pos]

                top_k.extend(top_k_words)

            # rank the top_k words by frequency
            top_k = [word for word in top_k if word not in stopwords.words('english')]
            top_k = collections.Counter(top_k)
            top_k = [word for word, _ in top_k.most_common(k)]
            return top_k
        
        except ValueError:
            print(f'The word: "{main_word}" does not exist in the list of tokens')
            return []


    def _cutDoc(
            self, 
            main_word: str, 
            doc: Optional[str] = None, 
            max_length: int = 512
            ) -> Optional[str]:

        main_word = ' ' + main_word.strip() + ' '
        if doc is None:
            doc = main_word
            return doc
        
        tokens = self.tokenizer.tokenize(doc)
        main_token = self.tokenizer.tokenize(main_word)[0]
       
        try:
            main_index = tokens.index(main_token)
        
        except ValueError:
            return None
        
        start = max(0, main_index - max_length//2)
        end = start + max_length

        if end > len(tokens):
            end = len(tokens)
            start = max(0, end - max_length)

        # Convert the tokens back to ids for decoding
        token_ids = self.tokenizer.convert_tokens_to_ids(tokens[start:end])

        # Decode the token ids back to a string
        sliced_doc = self.tokenizer.decode(token_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)

        return sliced_doc


def train_test_split(data: List[str], test_ratio=0.2, random_seed=None):
    """
    Split the data into train and test sets.

    Args:
        data (List[str]): The data to split.
        test_ratio (float): The ratio of the test set.
        random_seed (int): The random seed.

    Returns:
        data Tuple[List[str], List[str]]: The training and testing datasets.
    """
    
    if random_seed:
        random.seed(random_seed)
    data_copy = data[:]
    random.shuffle(data_copy)
    split_idx = int(len(data_copy) * (1 - test_ratio))
    train_data = data_copy[:split_idx]
    test_data = data_copy[split_idx:]
    return train_data, test_data



import pandas as pd
from sklearn.base import BaseEstimator
import numpy as np
from pathlib import Path
import pickle
from tqdm.autonotebook import trange, tqdm
from scipy.optimize import minimize
from .prr_time_independent_weights import PoissonReducedRankTimeIndependentWordWeights
from typing import Optional
from .structures import PoissonReducedRankParameters


class PenalizedPoissonReducedRankTimeDependentWordWeights(BaseEstimator):
    """This model is the poisson reduced rank model with time dependent wword weights. In this model, the word weights are allowed to depend on time t.
    The model is optimized in two steps. First, the model runs through an identification step in which the fussed lasso problem in eq. 5 is solved (constrained model/log-likelihood optimization).
    In the second step, the model is optimized in the estimation step (unconstrained log-likelihood) optimization.
    Note:
        - Currently, only balanced data is supported, i.e. each document is available for the same time indices t.
        - For now, the last term in eq. 5, involving the time lags of word weights, is not implemented. Optimization of the constrained model only happens on the first two terms in eq.5.
    """

    def __init__(self, I: int, T: int, J: int, K: int = 2, **kwargs):
        """Initializes a Penalized Reduced Rank Model with time dependent word weights.

        Args:
            K (int, optional): Number of word vector dimensions. Defaults to 2.
        """
        self.k = K
        self.I = I
        self.T = T
        self.J = J

        # define parameters
        self.alpha = np.empty((self.J))
        self.beta = np.empty((self.I * self.T))
        self.b = np.empty((self.k, self.J, self.T))
        self.f = np.empty((self.k, self.I, self.T))
        self.roh1 = np.empty((self.J, self.k))
        self.roh2 = np.empty((self.J, self.k))
        self.X: Optional[np.ndarray] = None

        ## initialize hyper parameters
        self.hyp_roh: Optional[np.ndarray] = None
        self.hyp_delta = None

        # params of the first starting model
        self.params_start_model: Optional[PoissonReducedRankParameters] = None

        # process possible hyperparamteres in **kwargs
        if "roh" in kwargs.keys():
            if isinstance(kwargs["roh"], np.ndarray) == False:
                raise ValueError(
                    "Keyword argument roh must be of type np.ndarray (i.e. created with np.array)"
                )
            if kwargs["roh"].shape != (K,):
                raise ValueError("roh must be of shape (K,)")
            if np.all(kwargs["roh"] > 0) == False:
                raise ValueError("All roh values must be greater 0")

            self.hyp_roh = kwargs["roh"]
        else:
            self.hyp_roh = np.full((self.k,), 0.5)

        if "delta" in kwargs.keys():
            if isinstance(kwargs["delta"], np.ndarray) == False:
                raise ValueError(
                    "Keyword argument delta must be of type np.ndarray (i.e. created with np.array)"
                )
            if kwargs["delta"].shape != (K,):
                raise ValueError("delta must be of shape (K,)")
            if np.all((kwargs["delta"] > 0) & (kwargs["delta"] < 1)) == False:
                raise ValueError("All delta values must be in the open interval (0,1)")

            self.hyp_delta = kwargs["delta"]
        else:
            self.hyp_delta = np.full((self.k,), 0.5)

    @staticmethod
    def calculate_IJT(
        texts: pd.DataFrame,
        text_column: str = "text",
        date_column: str = "date",
        individual_column: str = "individual",
    ) -> dict:
        """Calculates I, J, T for a given dataframe, which can be used for the model initialization.

        Args:
            texts (pd.DataFrame): Dataframe containing the data
            text_column (str, optional): column name containing the texts. Defaults to "text".
            date_column (str, optional): column name containing the date. Defaults to "date".
            individual_column (str, optional): column name containing the individual. Defaults to "individual".

        Returns:
            dict: Dictionary with keys I, T, J
        """
        stats = {}
        stats["I"] = len(texts.loc[:, individual_column].unique())
        stats["T"] = len(texts.loc[:, date_column].unique())

        texts_list = sum(texts[text_column].to_list(), [])
        vocab = list(dict.fromkeys(texts_list))
        stats["J"] = len(vocab)

        return stats

    @staticmethod
    def _compute_roh1_matrix(
        roh: np.ndarray, delta: np.ndarray, b: np.ndarray
    ) -> np.ndarray:
        """Computes the parameter matrix roh_{1,j}^k used in the fussed lasso optimization problem

        Args:
            roh (np.ndarray): hyper parameter used for the optimization of the fussed lasso problem. Must be of shape (k,)
            delta (np.ndarray): hyper parameter used for the optimization of the fussed lasso problem. Must be of shape (k,)
            b (np.ndarray): word weights of the poisson reduced rank method with independent word weights. Must be of the shape (J, k)

        Returns:
            np.ndarray: computed matrix for roh_1. same shape as input b.
        """
        numerator = roh * delta
        denominator = np.abs(b)
        return numerator / denominator

    @staticmethod
    def _compute_roh2_matrix(
        roh: np.ndarray, delta: np.ndarray, b: np.ndarray
    ) -> np.ndarray:
        """Computes the parameter matrix roh_{2,j}^k used in the fussed lasso optimization problem

        Args:
            roh (np.ndarray): hyper parameter used for the optimization of the fussed lasso problem. Must be of shape (k,)
            delta (np.ndarray): hyper parameter used for the optimization of the fussed lasso problem. Must be of shape (k,)
            b (np.ndarray): word weights of the poisson reduced rank method with independent word weights. Must be of the shape (J, K)

        Returns:
            np.ndarray: computed matrix for roh_2
        """
        numerator = roh * (1 - delta)
        denominator = np.abs(b)
        return numerator / denominator

    def _convert_params_2Dto3D(
        self, alpha: np.ndarray, beta: np.ndarray, b: np.ndarray, f: np.ndarray
    ) -> None:
        """Converts the parameters from the first step [S1], which are in 2D matrices to 3D parameter matrices

        Args:
            alpha (np.ndarray): array containing a_j
            beta (np.ndarray): array containing b_it
            b (np.ndarray): matrix of the form (J_tokens,K)
            f (np.ndarray): matrix of the form (K, L_documents)

        Returns:
            None: Overrides the class intern parameters to 3D matrices
        """
        self.alpha = alpha
        self.beta = beta
        self.b = (
            b.transpose().reshape(self.k, self.J, 1).repeat(repeats=self.T, axis=2)
        )  # (K,J,T)
        self.f = f.transpose().reshape(self.k, self.I, self.T)  # tranpose(0,2,1)

    @staticmethod
    def get_log_likelihood(y: np.ndarray, mu: np.ndarray) -> np.ndarray:
        """Retrieves the log likelihood matrix, where each element is computed by -mu + y*log(mu)

        Args:
            y (np.ndarray): matrix containing the real token counts. Expected shape (J, I*T)
            mu (np.ndarray): matrix containing the mu values. Expected shape (J, I*T)

        Returns:
            np.ndarray: matrix containing the computed log-likelihood for each element. The total sum of can be computed via np.sum
        """
        assert y.shape == mu.shape, "y and mu must be of same shape"
        return -mu + np.multiply(y, np.log(mu))

    def _fussed_lasso_computation(
        self, roh: np.ndarray, matrix: np.ndarray
    ) -> np.float32:
        """Function can be used for the fussed lasso terms of the model

        Args:
            roh (np.ndarray): parameter matrix of the the form (J,K)
            matrix (np.ndarray): matrix to compute the sums over. Must be of shape (K,J,T).

        Returns:
            np.float32: float containing the sum of all matrix elements.
        """
        assert roh.shape == (matrix.shape[1], self.k), "roh mst be of shape (J,K)"
        # reshape roh to get (K,J,1)
        roh = roh[..., np.newaxis].transpose(1, 0, 2)
        roh_matrix = np.multiply(roh, np.abs(matrix))
        sum_jt = np.sum(roh_matrix, axis=(1, 2))
        sum_jt_k = np.sum(sum_jt)
        return sum_jt_k

    def _obj_function_penalized(
        self,
        y: np.ndarray,
        roh1: np.ndarray,
        roh2: np.ndarray,
        alpha: np.ndarray,
        beta: np.ndarray,
        b: np.ndarray,
        f: np.ndarray,
    ) -> np.float32:
        """Computes the sum of the loss function for the fussed lasso problem given in formula (5) of the paper.

        Args:
            y (np.ndarray): matrix of y token counts. Must be of the form (J, T*I)
            roh1 (np.ndarray): roh1 matrix used for the first term. Must be of the form (J,K)
            roh2 (np.ndarray): roh matrix used for the second (time-differenced) term. Must be of the form (J,K)
            alpha (np.ndarray): array with J elements
            beta (np.ndarray): array with I*T elements
            b (np.ndarray): b matrix containing b_jt^(k). Must be of the form (K,J,T).
            f (np.ndarray): f matrix containing b_it^(k). Must be of the form (K,I,T).

        Returns:
            np.float32: positive Value of the computed loss function.
        """

        mu = self.get_mu(b=b, f=f, alpha=alpha, beta=beta)
        neg_ll = -1 * np.sum(self.get_log_likelihood(y, mu))
        # calculate fussed lasso terms
        term1 = self._fussed_lasso_computation(roh1, b)
        # time differenced term
        b_diff = np.diff(b, n=1, axis=2)  # axis 2 is the time axis.
        term2 = self._fussed_lasso_computation(roh2, b_diff)

        # calculate penalized part
        return neg_ll + term1 + term2

    @staticmethod
    def get_mu(
        b: np.ndarray,
        f: np.ndarray,
        alpha: Optional[np.ndarray] = None,
        beta: Optional[np.ndarray] = None,
    ):
        """Uses the input parameters of the model to compute the mu matrix for the Poisson Matrix

        Args:
            b (np.ndarray): parameter matrix containing b_jt. Must be of form (K, J_tokens, T_timepoints)
            f (np.ndarray): parameter matrix containing f_it. Must be of dimensions (K, I_documents, T_timepoints)
            alpha (Optional[np.ndarray], optional): Vector containing a_j. Must be of dimensions (J_tokens,). Defaults to None.
            beta (Optional[np.ndarray], optional): Vector containing b_it. Must be of dimensions (I*T,) Defaults to None.

        Returns:
            np.ndarray: returns an np.ndarray of the form (J,T*I), where the first T columns correspond to T = 1 for all individuals
        """
        assert b.shape[2] == f.shape[2], "Time dimensions for b and f not equal"
        assert b.shape[0] == f.shape[0], "Dimension k not equal for b and f"

        res = []
        for i in range(b.shape[2]):
            # print(i)
            res.append(np.tensordot(b[:, :, i], f[:, :, i], axes=((0), (0))))
            # print(np.tensordot(b[:,:,i],f[:,:,i],axes=((0),(0))))

        concat_array = np.concatenate(res, axis=1)
        # add alpha
        if alpha is not None:
            concat_array = np.add(alpha.reshape(-1, 1), concat_array)
        if beta is not None:
            concat_array = np.add(beta.reshape(1, -1), concat_array)

        return concat_array

    def fit(
        self,
        texts: pd.DataFrame,
        text_column: str = "text",
        date_column: str = "date",
        individual_column: str = "individual",
        n_iter: int = 100,
        train_params: dict = {"prrr_indep_niter": 5},
        tol: float = 1e5,
        warm_start: Optional[PoissonReducedRankParameters] = None,
    ):
        """Fits a Poisson Reduced Rank model with time dependent word weigths to the given term-document-matrix (tdm). The optimization routine implemented in this paper is described in Jentsch et. al. 'Time-dependent Poisson reduced rank models for political text data analysis' section 2.3.
            For the unpenalized/unconstrained model, please see the class UnconstrainedPoissonReducedRankTimeDependentWordWeights.

        Args:
            texts (pd.DataFrame): A pandas DataFrame containing the columns text_column, date_column and individual_column containing the documents and their respective dates. The column combination of individual_column and date_column leads to a panel like data structure.
                   The dates must be in a format interpretable by pandas.to_datetime(). Each element of the text_column column must be a list of strings. Internally, the create_dtm function is used with no control over the vocab argument of the create_dtm function currently.
            text_column (str): The name of the column in texts containing the documents. Defaults to "text".
            date_column (str): The name of the column in texts containing the dates. Defaults to "date".
            individual_column (str): An id column, containing the ids (ints) of the individual. Defaults to "individual".
            n_iter (int, optional): Maximum number of iterations for the optimization of the model. Defaults to 100.
            tol (float, optional): Currently not implemented. Tolerance for no improvement. If the difference between the likelihoods of the previous and last iteration is smaller tol, the optimization is stopped. Defaults to 1e5.
            train_params (dict): Additional parameters for the training process. prrr_indep_niter controlls the number of iterations for the independent word weights model used to as a starting point. prrr_indep_niter must be present.
        """
        # TODO: Implement possibility that parameters from previous model can be used.
        self.df = texts.copy()

        if warm_start:
            print("Using params from warm start...")
            current_params = warm_start
            model = PoissonReducedRankTimeIndependentWordWeights(self.k)

        else:
            print("Calculate initial values for the estimation...")

            # [S1] get initial values for the estimation
            model = PoissonReducedRankTimeIndependentWordWeights(self.k)

            current_params = model.fit(
                texts,
                text_column=text_column,
                date_column=date_column,
                individual_column=individual_column,
                n_iter=train_params["prrr_indep_niter"],
            )

        self.X = model._df2dtm(
            texts,
            text_column=text_column,
            date_column=date_column,
            individual_column=individual_column,
        )  # returns (J_tokens, L_documents) dtm matrix

        self.params_start_model = current_params

        print("Convert initial parameters to right dimensions")
        # convert the parameters from the poisson reduced rank model to the params of this models.
        # In particular f must be reshaped and the parameters b_j must be duplicated for all time points t.
        self._convert_params_2Dto3D(
            current_params.alpha,
            current_params.beta,
            current_params.b,
            current_params.f,
        )

        # compute roh 1 and roh 2.
        self.roh1 = self._compute_roh1_matrix(
            self.hyp_roh, self.hyp_delta, current_params.b
        )
        self.roh2 = self._compute_roh2_matrix(
            self.hyp_roh, self.hyp_delta, current_params.b
        )

        print("Optimize model...")
        # [S2] Repeat (i) - (v) below until a convergence criterion is met for steps m=1,....,M
        for i in trange(n_iter):
            # (i) set current params from the intial estimation (ommitted here, just use current_params)

            # (ii) fix current values alpha_j and b_jt and update beta and f by maximizing the poisson log likelihood (done internally)
            self._update_f_and_beta()

            #  (iii) fix the beta and f from from the previous run/inital params and update alpha and b solving for the fused lasso problem (see formula 5 in paper)
            self._update_b_and_alpha()

            # (iv) + (v) Reparameterize and override the current params
            self._reparameterize()

        return current_params

    def _reparameterize(self):
        """Reparameterizes the variables beta, alpha, b and f  according to step [S2] (iv)"""
        self.alpha = self.alpha + np.sum(self.beta) / self.I * self.T
        self.beta = self.beta - np.sum(self.beta) / self.I * self.T
        self.b = self.b * np.sqrt(np.sum(self.f**2) / self.I * self.T)
        self.f = self.f / np.sqrt(np.sum(self.f**2) / self.I * self.T)

    def _update_f_and_beta(self):
        """Updates the parameter vectors f and beta for the identification of th emodel

        Args:
            y (np.ndarray): y token matrix of the form (J,I*T)
        """

        def wrapper_optim_f_and_beta_given_alpha_and_b(
            beta_and_f: np.ndarray, *args
        ) -> np.float32:
            """Function used as a wrapper function in scipy optimize. Since scipy.optimize only takes 1D arrays

            Args:
                beta_and_f (np.ndarray): flattened np.ndarray. The first I*T values correspond to the vector beta and the remaining values to f. F is reshaped for computation of the loss.
                *args (Tuple): Tuple of the form (y,alpha,b, len_beta, K, I, J, T) the values for y, alpha and b. No reshaping is needed, just pass them in a tuple as they are

            Returns:
                np.float32: Loss of negative log likelihood which is minimized be the scipy.minimize function. Just the matrix sum of the log-likelihood matrix.
            """

            # unpack *args arguments (y,alpha,b) expected
            y = args[0]
            alpha = args[1]
            b = args[2]
            len_beta = args[3]
            K = args[4]
            I = args[5]
            J = args[6]
            T = args[7]

            # get beta
            beta1D = beta_and_f[:len_beta]
            f1D = beta_and_f[len_beta:]
            # reshape f to get matrix for get_mu
            f = f1D.reshape(K, I, T)

            # note: since the wrapper function is enclosed by a function with acces to self, I do not need to pass self
            mu = self.get_mu(b=b, f=f, alpha=alpha, beta=beta1D)
            return -1 * np.sum(self.get_log_likelihood(y, mu))

        ## flatten current params
        f_lattened = self.f.flatten(order="C")
        beta = self.beta.flatten()
        len_beta = beta.size

        concat_beta_and_f = np.concatenate([beta, f_lattened], axis=0)

        optim_result = minimize(
            wrapper_optim_f_and_beta_given_alpha_and_b,
            x0=concat_beta_and_f,
            args=(self.X, self.alpha, self.b, len_beta, self.k, self.I, self.J, self.T),
        )

        # unpack results from the optimization
        new_beta, new_f = optim_result.x[:len_beta], optim_result.x[len_beta:]

        # update internal params with new values
        self.f = new_f.reshape(self.f.shape)
        self.beta = new_beta.reshape(self.beta.shape)

    def _update_b_and_alpha(self):
        """Updates the internal parameters alpha and b given f and beta

        Args:
            y (np.ndarray): token matrix of the form (J, I*T)
        """

        def wrapper_optim_b_and_alpha_given_beta_and_f(
            alpha_and_b: np.ndarray, *args
        ) -> np.float32:
            """Function used as a wrapper function in scipy optimize. Since scipy.optimize only takes 1D arrays

            Args:
                alpha_and_b (np.ndarray): flattened np.ndarray. The first J values correspond to the vector alpha and the remaining values to the flattend b. b is reshaped for computation of the loss.
                *args (Tuple): Tuple of the form (y,beta,f, len_alpha, K, I, J, T, roh1, roh2). No reshaping is needed as it is done internally, just pass them in the tuple as they are.

            Returns:
                np.float32: Value for the penalized loss function to optimize for
            """

            # unpack *args arguments (y,alpha,b) expected
            y = args[0]
            beta = args[1]
            f = args[2]
            len_alpha = args[3]
            K = args[4]
            I = args[5]
            J = args[6]
            T = args[7]
            roh1 = args[8]
            roh2 = args[9]

            # get alpha
            alpha = alpha_and_b[:len_alpha]
            b1D = alpha_and_b[len_alpha:]
            # reshape f to get matrix for get_mu
            b = b1D.reshape(K, J, T)

            # note: since the wrapper function is enclosed by a function with acces to self, I do not need to pass self

            loss = self._obj_function_penalized(
                y=y, roh1=roh1, roh2=roh2, alpha=alpha, beta=beta, b=b, f=f
            )
            return loss

        ## flatten current params
        b_flattened = self.b.flatten(order="C")
        alpha = self.alpha.flatten()
        len_alpha = alpha.size

        concat_beta_and_f = np.concatenate([alpha, b_flattened], axis=0)

        optim_result = minimize(
            wrapper_optim_b_and_alpha_given_beta_and_f,
            x0=concat_beta_and_f,
            args=(
                self.X,
                self.beta,
                self.f,
                len_alpha,
                self.k,
                self.I,
                self.J,
                self.T,
                self.roh1,
                self.roh2,
            ),
        )

        # unpack results from the optimization
        new_alpha, new_b = optim_result.x[:len_alpha], optim_result.x[len_alpha:]

        # override internal params with updated values
        self.b = new_b.reshape(self.b.shape)
        self.alpha = new_alpha.reshape(self.alpha.shape)

    def save(self, path: Path) -> None:
        """Saves the Poisson Reduced rank model with independent word weights to the given path
        Args:
            path (Path): Path object to the given save location
        """
        with open(path.absolute(), "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: Path) -> dict:

        with open(path.absolute(), "rb") as f:
            return pickle.load(f)

    def fit_update(self):
        raise NotImplementedError

    def infer_vector(self):
        raise NotImplementedError

    def plot_word_weights(self, params: PoissonReducedRankParameters):
        raise NotImplementedError

    def get_params(self) -> dict:
        """Returns the parameters for the poisson reduced rank model with independent word weights

        Returns:
            dict: Dictionary containing the parameter arrays
        """
        return {
            "K": self.k,
            "alpha": self.alpha,
            "beta": self.beta,
            "b": self.b,
            "f": self.f,
        }

    def set_params(self, **params) -> None:
        """sets the internal parameters of the poisson reduced rank model. The **params should contains 'alpha', 'beta', 'b' and 'f'. 'K' is not allowed, since it would change the whole metho due to the increase in dimensions.
        The dimensionalities are not checked, please make sure that the np.ndarrays are of the correct shape.
        """

        if "alpha" in params:
            self.alpha = params["alpha"]

        if "beta" in params:
            self.beta = params["beta"]

        if "b" in params:
            self.b = params["b"]

        if "f" in params:
            self.f = params["f"]



from distutils.core import setup, Extension
import numpy
from Cython.Distutils import build_ext
import os

os.environ["C_INCLUDE_PATH"] = numpy.get_include()
setup(
    cmdclass={'build_ext': build_ext},
    ext_modules=[Extension("ttta.methods.LDA.lda_gibbs",
                 sources=["src/ttta/methods/LDA/lda_gibbs.pyx"],
                 include_dirs=[numpy.get_include()]),
                 ],
)



import pandas as pd
from typing import Union, List
from datetime import datetime
import warnings
import numpy as np
from ..methods.lda_prototype import LDAPrototype

def _get_time_indices(texts: pd.DataFrame, how: Union[str, List[datetime]] = "M", last_date: datetime = None,
                      date_column: str = "date", min_docs_per_chunk: int = 500) -> pd.DataFrame:
    """
        Creates the time indices for the given texts. If update is True, the time indices are appended to the existing ones.
        Args:
            texts: A pandas DataFrame containing the documents and their respective dates.
                   The dates must be in a format interpretable by pandas.to_datetime(). The texts must be a list of strings.
            how: List of datetime dates indicating the end of time chunks or a string indicating the frequency of the time chunks. Used to create time
                 chunks when fixed dates were used for the initial fit. If None, the same time chunk rule as in the initial fit is used.
            last_date: The last date of the texts that were used before. If None, the last date of the texts is used.
            date_column: The name of the column containing the dates in the texts DataFrame.
            min_docs_per_chunk: The minimum number of documents that a time chunk must contain. If a time chunk does not contain at least this many
                                documents, it is combined with the next time chunk.
    """
    if not isinstance(texts, pd.DataFrame):
        raise TypeError("texts must be a pandas DataFrame!")
    if not isinstance(date_column, str):
        raise TypeError("date_column must be a string!")
    if min_docs_per_chunk is None:
        min_docs_per_chunk = 1
    if not isinstance(min_docs_per_chunk, int):
        try:
            if min_docs_per_chunk == int(min_docs_per_chunk):
                min_docs_per_chunk = int(min_docs_per_chunk)
            else:
                raise ValueError
        except ValueError:
            raise TypeError("min_docs_per_chunk must be an integer!")
    if min_docs_per_chunk < 1:
        raise ValueError("min_docs_per_chunk must be a natural number greater than 0!")
    df_index = pd.DataFrame({'chunk_start': range(len(texts))}, index=pd.to_datetime(texts[date_column]))
    if last_date is not None:
        if df_index.index.min() <= last_date:
            raise ValueError("The dates of the new texts chronologically overlap with the text that were already used before!")

    if not isinstance(how, (str, list, np.ndarray)):
        raise TypeError("how must be a string or a list of datetime objects!")
    if isinstance(how, str):
        period_start = df_index.resample(how).first()
    else:
        if df_index.index.max() < how[-1]:
            how = how[:-1]
        if sum([df_index.index.min() > x for x in how]) > 1:
            how = how[(sum([df_index.index.min() > x for x in how]) - 1):]
        chunk_dates = pd.DatetimeIndex(how)
        chunks = chunk_dates.searchsorted(df_index.index, "right")
        period_start = df_index.groupby(chunk_dates[np.where(chunks < 1, 0, chunks - 1)]).first()
        period_start.index.name = date_column

    len_before_na_drop = len(period_start)
    period_start = period_start.dropna()
    rows_to_drop = []
    current_docs_since_last_chunk = 0
    for i, (index, row) in enumerate(period_start[1:].iterrows()):
        current_docs_since_last_chunk += row["chunk_start"] - period_start["chunk_start"].iloc[i]
        if current_docs_since_last_chunk < min_docs_per_chunk and not i == len(period_start) - 2:
            rows_to_drop.append(index)
        else:
            current_docs_since_last_chunk = 0
    period_start = period_start.drop(rows_to_drop)
    if len(period_start) < len_before_na_drop:
        warnings.warn(f"{len_before_na_drop - len(period_start)} time chunks do not fulfill the requirement of containing least {min_docs_per_chunk} texts and are combined "
                      f"into larger time chunks.")
    period_start["chunk_start"] = period_start["chunk_start"].astype(np.uint64)
    period_start = period_start.reset_index()
    return period_start



from typing import Union, List, Tuple, Set
import numpy as np
import pandas as pd
import matplotlib
from .rolling_lda import RollingLDA
import seaborn as sns
from matplotlib import pyplot as plt
import math
import pickle
import os
from scipy.spatial.distance import cosine
from operator import itemgetter

class TopicalChanges:
    def __init__(self, roll: RollingLDA, mixture: float = 0.9, reference_period: int = 4, quantile_threshold: float = 0.99, samples: int = 500, save_path: str = None,
                 load_path: str = None, fast: bool = True):
        """
            Calculates topical change points for the given parameters.
            Args:
                mixture: The mixture parameter as an a-priori-estimator for expected change in word usage from one time-chunk to the next. Lower values
                         will detect more change points, but the changes will be smaller. Higher values will detect fewer change points, but the changes
                         will be more noticeable.
                reference_period: The maximum number of chunks to look back for when calculating the topic proportions in the simulation. This can make
                                  the simulation more robust to outliers and allows for incorporating seasonal effects, e.g. when including the last four
                                  chunks when making a quarterly analysis. When using a high value, the change detection will detect gradual changes, since
                                  it compares the new chunks to very old chunks and when using a low value, the change detection will detect sudden changes,
                                  since it compares the new chunks to very recent chunks.
                quantile_threshold: The bootstrap quantile to compare the observed chunk similarities to, to detect significant changes in word usage.
                samples: The number of samples to draw for the simulation.
                save_path: The path to a directory in which the simulated topic proportions should be saved. If None, the simulated topic proportions are not saved.
                load_path: The path to a directory from which the simulated topic proportions should be loaded. If None, the simulated topic proportions are not loaded.
                fast: Whether the word impact should be calculated fast. If True, words that do not occur in the compared time chunks are ignored during
                      the word impact calculation. This should not change the results, unless "word_impact_number" is set to a high value compared to the
                      vocabulary size.
            Returns:
                If word_impact is True, a pandas DataFrame containing the word impact for each topic and change point. Else None.
        """
        if not isinstance(mixture, float):
            try:
                mixture = float(mixture)
            except ValueError:
                raise TypeError("mixture must be a float!")
        if mixture <= 0 or mixture >= 1:
            raise ValueError("mixture must be a float between 0 and 1!")
        if not isinstance(reference_period, int):
            try:
                if reference_period == int(reference_period):
                    reference_period = int(reference_period)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("reference_period must be an integer!")
        if reference_period < 1:
            raise ValueError("reference_period must be a natural number greater than 0!")
        if not isinstance(quantile_threshold, float):
            try:
                quantile_threshold = float(quantile_threshold)
            except ValueError:
                raise TypeError("quantile_threshold must be a float!")
        if quantile_threshold <= 0 or quantile_threshold >= 1:
            raise ValueError("quantile_threshold must be a float between 0 and 1!")
        if not isinstance(samples, int):
            try:
                if samples == int(samples):
                    samples = int(samples)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("samples must be an integer!")
        if samples < 1:
            raise ValueError("samples must be a natural number greater than 0!")
        if not isinstance(fast, bool):
            raise TypeError("fast must be a boolean!")
        if not isinstance(roll, RollingLDA):
            raise TypeError("roll must be a RollingLDA object!")

        self.mixture = mixture
        self._roll = roll
        self.reference_period = reference_period
        self.quantile_threshold = quantile_threshold
        self.samples = samples
        self.fast = fast
        self._simulate_changes(save_path, load_path)


    def plot_distances(self, **plot_args) -> None:
        """
            Plots the distances between the observed and simulated topic proportions.
            Args:
                **plot_args: Additional arguments to be passed to the matplotlib.pyplot.subplot function, e.g. nrows, ncols and figsize.
            Returns:
                None
        """
        if len([np.argwhere(self._distances_simulated.transpose()[x] < self._distances_observed.transpose()[x]).transpose(1, 0)[0].tolist() for x in
                          range(len(self._distances_simulated.transpose()))]) < 1:
            raise ValueError("No change points detected!")
        matplotlib.use('TkAgg')
        top = self._roll.top_words(chunk=None, number=3, importance=True, return_as_data_frame=False)
        headers = [f"{k}: {', '.join(top[k])}" for k in range(self._roll._K)]
        similarities_observed = 1 - self._distances_observed[self._roll._warmup + 1:, :]
        similarities_simulated = 1 - self._distances_simulated[self._roll._warmup + 1:, :]
        sns.set(style="darkgrid")
        if len(plot_args) == 0:
            plot_args = {"nrows": max(1, math.floor(self._roll._K / 5)), "ncols": 5, "figsize": (20, self._roll._K)}
        fig, axs = plt.subplots(**plot_args, sharex='col', sharey='row')
        axs = axs.flatten()
        for i in range(similarities_observed.shape[1]):
            sns.lineplot(x=self._roll.chunk_indices[self._roll._date_column].iloc[self._roll._warmup + 1:], y=similarities_observed[:, i], ax=axs[i], color='#4575b4')
            sns.lineplot(x=self._roll.chunk_indices[self._roll._date_column].iloc[self._roll._warmup + 1:], y=similarities_simulated[:, i], ax=axs[i], color='#db9191')

            higher_points = np.where(similarities_simulated[:, i] > similarities_observed[:, i])[0]
            for point in higher_points:
                axs[i].axvline(x=self._roll.chunk_indices[self._roll._date_column].iloc[self._roll._warmup + 1:].iloc[point], color='red', linestyle='--', alpha=0.5)

            axs[i].set_ylabel("Similarities")
            axs[i].set_title(headers[i])
            axs[i].set_xlabel("Date")
            axs[i].tick_params(axis='x', rotation=45)
        plt.tight_layout()
        plt.suptitle("Topical Changes", fontsize=16, y=1.02)
        plt.show()

    def word_impact(self, number: int = 5, date_format: str = "%Y-%m-%d", fast: bool = True) -> pd.DataFrame:
        """
            Calculates the leave one out word impact for each topic and change point.
            Args:
                number: The number of words to return for the word impact.
                date_format: The date format to return alongside the changes.
                fast: Whether the word impact should be calculated fast. If True, words that do not occur in the compared time chunks are ignored during
                      the word impact calculation. This should not change the results, unless "number" is set to a high value compared to the
                      vocabulary size.
            Returns:
                A pandas DataFrame containing the date and word impact for each topic and change point.
        """
        if not isinstance(number, int):
            try:
                if number == int(number):
                    number = int(number)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("number must be an integer!")
        if number < 1:
            raise ValueError("number must be a natural number greater than 0")
        if not isinstance(date_format, str):
            raise TypeError("date_format must be a string!")
        if not isinstance(fast, bool):
            raise TypeError("fast must be a boolean!")

        assignments = np.array([self._roll.get_word_topic_matrix(chunk=chunk).transpose() for chunk, row in self._roll.chunk_indices.iterrows()])
        topics = assignments.transpose((1, 2, 0))
        change_indices = [np.argwhere(self._distances_simulated.transpose()[x] < self._distances_observed.transpose()[x]).transpose(1, 0)[0].tolist() for x in
                          range(len(self._distances_simulated.transpose()))]
        events_end_date = [[self._roll.chunk_indices[self._roll._date_column].iloc[y] for y in x] for x in change_indices]
        leave_one_out_word_impact = {"Topic": [], "Date": [], "Significant Words": []}
        for k, changes in enumerate(change_indices):
            if len(changes) == 0:
                continue
            for i, change in enumerate(changes):
                temp = topics[k, :, max(1, change - self.runs[change][k], change - self.reference_period):change].sum(axis=1)
                loo_temp = np.array(
                    [0 if fast and topics[k, word, change] == 0 and temp[word] == 0 else cosine(np.delete(topics[k, :, change], word), np.delete(temp, word)) for word in range(topics.shape[1])])
                leave_one_out_substracted = loo_temp - cosine(topics[k, :, change], temp)
                significant_words = itemgetter(*leave_one_out_substracted.argsort()[-number:].astype(np.uint64).tolist())(self._roll.lda.get_vocab())
                leave_one_out_word_impact["Topic"].append(k)
                leave_one_out_word_impact["Date"].append(events_end_date[k][i].strftime(date_format))
                leave_one_out_word_impact["Significant Words"].append(significant_words)
        leave_one_out_word_impact = pd.DataFrame(leave_one_out_word_impact)
        # todo loo richtig?
        return leave_one_out_word_impact

    def _simulate_changes(self, save_path: str = None, load_path: str = None) -> Union[Tuple[np.ndarray, np.ndarray], None]:
        """
            Simulates the topic proportions for each chunk and calculates the distances between the observed and simulated topic proportions.
            Args:
                save_path: The path to a directory in which the simulated topic proportions should be saved. If None, the simulated topic proportions are not saved.
                load_path: The path to a directory from which the simulated topic proportions should be loaded. If None, the simulated topic proportions are not loaded.
            Returns:
                The distances between the observed and simulated topic proportions.
        """
        if (not isinstance(save_path, str) and save_path is not None) or (not isinstance(load_path, str) and load_path is not None):
            raise TypeError("save_path and load_path must be string or None!")
        distances_simulated = np.zeros((len(self._roll.chunk_indices), self._roll._K))
        assignments = np.array([self._roll.get_word_topic_matrix(chunk=chunk).transpose() for chunk, row in self._roll.chunk_indices.iterrows()])
        topics = assignments.transpose((1, 2, 0))

        if load_path is not None:
            self._distances_observed = pd.read_csv(os.path.join(load_path, "observations.csv")).to_numpy()
            for chunk in range(self._roll._warmup + 1, len(self._roll.chunk_indices)):
                chunk_simulated = pd.read_csv(os.path.join(load_path, f"simulation_{chunk}.csv"))
                distances_simulated[chunk, :] = np.quantile(chunk_simulated, self.quantile_threshold, axis=0)
            self._distances_simulated = distances_simulated
            return

        def calculate_phi(x: np.ndarray) -> float: return (x + self._roll._gamma) / (np.sum(x, axis=1)[:, np.newaxis] + x.shape[1] * self._roll._gamma)
        phi_chunks = np.array([calculate_phi(x) for x in assignments])
        distances_observed = np.zeros((len(phi_chunks), self._roll._K))
        run_length = np.array([min(self._roll._warmup, self.reference_period) for x in range(self._roll._K)])
        self.runs = [run_length.copy()]
        for i, chunk in enumerate(phi_chunks):
            chunk_simulated = np.zeros((self.samples, self._roll._K))
            if i <= self._roll._warmup:
                self.runs.append(run_length.copy())
                continue
            lookback_by_topic = [min(run_length[x], self.reference_period) for x in range(self._roll._K)]
            topic_frequencies = assignments[i].sum(axis=1)
            for topic in range(self._roll._K):
                topics_run = topics[topic, :, max(1, i - lookback_by_topic[topic]):i].sum(axis=1)
                distances_observed[i, topic] = cosine(topics[topic, :, i], topics_run)
                topics_tmp = assignments[max(1, i - lookback_by_topic[topic]):i, :, :].sum(axis=0) + self._roll._gamma
                phi = topics_tmp / np.sum(topics_tmp, axis=1)[:, np.newaxis]
                topics_tmp = assignments[i] + self._roll._gamma
                phi_tmp = phi[topic, :].copy()
                phi = topics_tmp / np.sum(topics_tmp, axis=1)[:, np.newaxis]
                phi_tmp = (1 - self.mixture) * phi_tmp + self.mixture * phi[topic, :]
                simulations = self._sample(phi_tmp, topic_frequencies[topic], topics_run)
                chunk_simulated[:, topic] = simulations
                distances_simulated[i, topic] = np.quantile(simulations, self.quantile_threshold)
            run_length = np.where(distances_simulated[i, :] < distances_observed[i, :], 1, run_length + 1)
            self.runs.append(run_length.copy())
            if save_path is not None:
                if not os.path.exists(save_path):
                    os.makedirs(save_path)
                pd.DataFrame(chunk_simulated).to_csv(os.path.join(save_path, f"simulation_{i}.csv"), index=False)
        if save_path is not None:
            pd.DataFrame(distances_observed).to_csv(os.path.join(save_path, f"observations.csv"), index=False)
        self._distances_simulated = distances_simulated
        self._distances_observed = distances_observed
        return distances_simulated, distances_observed

    def _sample(self, phi, frequency, topics_run) -> List[float]:
        """
            Samples the topic proportions for the simulation.
            Args:
                phi: The topic proportions of the last chunks.
                frequency: The number of words in the current time chunk assigned to the topic to be simulated.
                topics_run: The topic proportions of the last chunk.
            Returns:
                The sampled topic proportions.
        """
        if not isinstance(phi, np.ndarray):
            raise TypeError("phi must be a numpy array!")
        if not isinstance(frequency, int):
            try:
                if frequency == int(frequency):
                    frequency = int(frequency)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("frequency must be an integer!")
        if frequency < 1:
            raise ValueError("frequency must be a natural number greater than 0")
        if not isinstance(topics_run, np.ndarray):
            raise TypeError("topics_run must be a numpy array!")
        distances = []
        sample = np.random.choice(len(self._roll.lda.get_vocab()), frequency * self.samples, p=phi)
        sample = sample.reshape((self.samples, frequency))
        vocab_size = len(self._roll.lda.get_vocab())
        for i in range(self.samples):
            unique, tmp_counts = np.unique(sample[i, :], return_counts=True)
            counts = np.zeros(vocab_size)
            counts[unique] = tmp_counts
            distances.append(cosine(counts, topics_run))
        return distances

    def save(self, path: str) -> None:
        """
        Saves the model to a pickle file
        Returns:
            None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        pickle.dump(self, open(path, "wb"))

    def load(self, path: str) -> None:
        """
        Loads a model from a pickle file
        Returns:
            None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        loaded = pickle.load(open(path, "rb"))
        self.__dict__ = loaded.__dict__



import random
from src.ttta.diachronic_embeddings.utils.components import WordSimilarities
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import BSpline, splrep
from typing import Literal
import warnings
from src.ttta.diachronic_embeddings.utils.settings import EmbeddingFiles, FileLoader
import json


class PolynomialFitting:
    def __init__(
            self,
            word:str
        ):
        self.files = EmbeddingFiles()
        self.word_proportions = []
        self.word = word
        self.years = self.files.years_used

        self.word_props, self.words = FileLoader.load_files(self.__class__.__name__, word)

        with open(f'../embeddings_similarity/embeddings_sim_w_.json') as f:
            self.word_props = json.load(f)

        for w_ in self.word_props:
             self.word_proportions += [WordSimilarities(**w_)]

        self.num_senses = len(self.word_proportions[0].props)


    def sense_distribution(self, sense_idx):
        if not sense_idx in range(0, self.num_senses + 1):
            raise ValueError(
                f'The sense index {sense_idx} not present in the range of senses available for the word {self.word}'
            )
        return [word_prop.props[sense_idx] for word_prop in self.word_proportions]

    def polynomial_fit(self, sense: int, deg:int=20):
        warnings.filterwarnings('ignore')
        dist = self.sense_distribution(sense)
        return np.poly1d(np.polyfit(self.years, dist, deg))

    def bspline_fit(self, sense:int):
        dist = self.sense_distribution(sense)
        tck_spline_args = splrep(self.years, dist, s=0, k=3)
        return BSpline(*tck_spline_args)

    def distribution_all_senses(self, fit:Literal['polynomial', 'bspline']):
        all_senses = []
        sense_ = {}
        xp = np.linspace(1980, 2018, 100)
        for sense_num in range(0, self.num_senses):
            sense_['years'] = self.years
            sense_['distribution'] = self.sense_distribution(sense_num)
            sense_['sense_id'] = sense_num

            if fit == 'polynomial':
                sense_['y_fit'] = self.polynomial_fit(sense_num)

            if fit == 'bspline':
                sense_['y_fit'] = self.bspline_fit(sense_num)(xp)

            all_senses += [sense_.copy()]

        return all_senses

def plot_word(word:str, fit:Literal['polynomial', 'bspline']):
    if not fit in ['polynomial', 'bspline']:
        raise ValueError(
            f'The fit type provided is not correct, expected "polynomial" or "bspline", got {type(fit)}'
        )

    fig, ax = plt.subplots()
    markers = ['o', 'v', '^', 's', 'p', 'P', 'h', 'H', 'D']
    random.shuffle(markers)
    poly_w1 = PolynomialFitting(word=word)
    xp = np.linspace(1980, 2018, 100)

    if fit == 'polynomial':
        distr_all_senses = poly_w1.distribution_all_senses(fit='polynomial')
        for sense, obj in zip(distr_all_senses, markers[:poly_w1.num_senses]):
            ax.plot(sense['years'], sense['distribution'], f'{obj}', label=f'{word}, for the sense: {sense["sense_id"]}')
            ax.plot(xp, sense['y_fit'](xp), '-')

    if fit == 'bspline':
        distr_all_senses = poly_w1.distribution_all_senses(fit='bspline')
        for sense, obj in zip(distr_all_senses, markers[:poly_w1.num_senses]):
            ax.plot(sense['years'], sense['distribution'], f'{obj}', label=f'{word}, for the sense: {sense["sense_id"]}')
            ax.plot(xp, sense['y_fit'], '-')

    ax.title.set_text(f'Word: {word}')
    ax.legend()
    plt.show()

def plot_words(words:tuple, sense_id_w1:int, sense_id_w2:int, sense_id_w3:int, fit:Literal['polynomial', 'bspline']):
    w_1, w_2, w_3= words

    poly_w1 = PolynomialFitting(word=w_1)
    poly_w2 = PolynomialFitting(word=w_2)
    poly_w3 = PolynomialFitting(word=w_3)

    xp = np.linspace(1980, 2018, 100)
    fig, ax = plt.subplots()
    dist_1 = poly_w1.sense_distribution(sense_idx=sense_id_w1)
    dist_2 = poly_w2.sense_distribution(sense_idx=sense_id_w2)
    dist_3 = poly_w3.sense_distribution(sense_idx=sense_id_w3)

    if fit == 'polynomial':
        p_1 = poly_w1.polynomial_fit(sense=sense_id_w1, deg=20)
        p_2 = poly_w2.polynomial_fit(sense=sense_id_w2, deg=20)
        p_3 = poly_w3.polynomial_fit(sense=sense_id_w3, deg=20)

        ax.plot(poly_w1.years, dist_1, '*', label=f'{w_1}, for the sense: {sense_id_w1}')
        ax.plot(xp, p_1(xp), '-', )
        ax.plot(poly_w1.years, dist_2, '*', label=f'{w_2} for the sense: {sense_id_w2}')
        ax.plot(xp, p_2(xp), '-',)

        ax.plot(poly_w1.years, dist_3, '+', label=f'{w_3} for the sense: {sense_id_w3}')
        ax.plot(xp, p_3(xp), '-', )

    if fit == 'bspline':
        b_1 = poly_w1.bspline_fit(sense=sense_id_w1)
        b_2 = poly_w2.bspline_fit(sense=sense_id_w2)
        b_3 = poly_w3.bspline_fit(sense=sense_id_w3)

        ax.plot(poly_w1.years, dist_1, '*', label=f'{w_1}, for the sense: {sense_id_w1}')
        ax.plot(xp, b_1(xp), '-', )
        ax.plot(poly_w1.years, dist_2, '*', label=f'{w_2} for the sense: {sense_id_w2}')
        ax.plot(xp, b_2(xp), '-', )

        ax.plot(poly_w1.years, dist_3, '+', label=f'{w_3} for the sense: {sense_id_w3}')
        ax.plot(xp, b_3(xp), '-', )


    ax.legend()
    plt.show()


if __name__ == '__main__':
    # plot_words(('abuse', 'black', 'kill'), sense_id_w1=2, sense_id_w2=2, sense_id_w3=2, fit='bspline')
    # plot_word('abuse', fit='bspline')
    p = PolynomialFitting(word='abuse')
    print(p.distribution_all_senses(fit='bspline'))